@@ -1,4 +1,4 @@
-/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2019 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,8 +11,8 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-*/
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ */
 
 #pragma once
 
@@ -30,7 +30,7 @@
 typedef struct {
 	knot_time_t now;
 
-	kasp_db_t **kasp_db;
+	knot_lmdb_db_t *kasp_db;
 	knot_kasp_zone_t *zone;
 	knot_kasp_policy_t *policy;
 	dnssec_keystore_t *keystore;
@@ -38,6 +38,10 @@ typedef struct {
 	char *kasp_zone_path;
 
 	bool rrsig_drop_existing;
+	bool keep_deleted_keys;
+	bool keytag_conflict;
+
+	knot_rrset_t *offline_rrsig;
 } kdnssec_ctx_t;
 
 /*!
@@ -46,10 +50,11 @@ typedef struct {
  * \param conf         Configuration.
  * \param ctx          Signing context to be initialized.
  * \param zone_name    Name of the zone.
+ * \param kaspdb       Key and signature policy database.
  * \param from_module  Module identifier if initialized from a module.
  */
 int kdnssec_ctx_init(conf_t *conf, kdnssec_ctx_t *ctx, const knot_dname_t *zone_name,
-                     const conf_mod_id_t *from_module);
+                     knot_lmdb_db_t *kaspdb, const conf_mod_id_t *from_module);
 
 /*!
  * \brief Save the changes in ctx (in kasp zone).
