@@ -11,7 +11,7 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #include <assert.h>
@@ -354,7 +354,8 @@ int knot_check_signature(const knot_rrset_t *covered,
                     const knot_rrset_t *rrsigs, size_t pos,
                     const dnssec_key_t *key,
                     dnssec_sign_ctx_t *sign_ctx,
-                    const kdnssec_ctx_t *dnssec_ctx)
+                    const kdnssec_ctx_t *dnssec_ctx,
+                    bool skip_crypto)
 {
 	if (knot_rrset_empty(covered) || knot_rrset_empty(rrsigs) || !key ||
 	    !sign_ctx || !dnssec_ctx) {
@@ -364,11 +365,17 @@ int knot_check_signature(const knot_rrset_t *covered,
 	knot_rdata_t *rrsig = knot_rdataset_at(&rrsigs->rrs, pos);
 	assert(rrsig);
 
-	if (is_expired_signature(rrsig, dnssec_ctx->now,
-	                         dnssec_ctx->policy->rrsig_refresh_before)) {
+	// consider signature invalid even if validity ends in refresh - in order to refresh it soon enough
+	knot_timediff_t refresh = dnssec_ctx->policy->rrsig_refresh_before +
+	                          dnssec_ctx->policy->rrsig_prerefresh;
+	if (is_expired_signature(rrsig, dnssec_ctx->now, refresh)) {
 		return DNSSEC_INVALID_SIGNATURE;
 	}
 
+	if (skip_crypto) {
+		return KNOT_EOK;
+	}
+
 	// identify fields in the signature being validated
 
 	dnssec_binary_t signature = {
