@@ -11,20 +11,24 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #include <assert.h>
+#include <pthread.h>
 #include <sys/types.h>
 
 #include "libdnssec/error.h"
 #include "libdnssec/key.h"
 #include "libdnssec/keytag.h"
 #include "libdnssec/sign.h"
+#include "knot/common/log.h"
 #include "knot/dnssec/key-events.h"
+#include "knot/dnssec/key_records.h"
 #include "knot/dnssec/rrset-sign.h"
 #include "knot/dnssec/zone-sign.h"
 #include "libknot/libknot.h"
+#include "contrib/dynarray.h"
 #include "contrib/macros.h"
 #include "contrib/wire_ctx.h"
 
@@ -70,7 +74,7 @@ static bool apex_rr_changed(const zone_node_t *old_apex,
 	knot_rrset_t old_rr = node_rrset(old_apex, type);
 	knot_rrset_t new_rr = node_rrset(new_apex, type);
 
-	return !knot_rrset_equal(&old_rr, &new_rr, KNOT_RRSET_COMPARE_WHOLE);
+	return !knot_rrset_equal(&old_rr, &new_rr, false);
 }
 
 static bool apex_dnssec_changed(zone_update_t *update)
@@ -94,6 +98,8 @@ static bool apex_dnssec_changed(zone_update_t *update)
  * \param key      Signing key.
  * \param ctx      Signing context.
  * \param policy   DNSSEC policy.
+ * \param skip_crypto All RRSIGs in this node have been verified, just check validity.
+ * \param at       RRSIG position.
  *
  * \return The signature exists and is valid.
  */
@@ -101,7 +107,9 @@ static bool valid_signature_exists(const knot_rrset_t *covered,
 				   const knot_rrset_t *rrsigs,
 				   const dnssec_key_t *key,
 				   dnssec_sign_ctx_t *ctx,
-				   const kdnssec_ctx_t *dnssec_ctx)
+				   const kdnssec_ctx_t *dnssec_ctx,
+                                   bool skip_crypto,
+				   uint16_t *at)
 {
 	assert(key);
 
@@ -122,7 +130,10 @@ static bool valid_signature_exists(const knot_rrset_t *covered,
 		}
 
 		if (knot_check_signature(covered, rrsigs, i, key, ctx,
-		                         dnssec_ctx) == KNOT_EOK) {
+		                         dnssec_ctx, skip_crypto) == KNOT_EOK) {
+			if (at != NULL) {
+				*at = i;
+			}
 			return true;
 		}
 	}
@@ -153,7 +164,7 @@ static bool all_signatures_exist(const knot_rrset_t *covered,
 
 		if (!valid_signature_exists(covered, rrsigs, key->key,
 		                            sign_ctx->sign_ctxs[i],
-		                            sign_ctx->dnssec_ctx)) {
+		                            sign_ctx->dnssec_ctx, false, NULL)) {
 			return false;
 		}
 	}
@@ -170,110 +181,43 @@ static bool all_signatures_exist(const knot_rrset_t *covered,
 static void note_earliest_expiration(const knot_rdata_t *rrsig, knot_time_t *expires_at)
 {
 	assert(rrsig);
-	assert(expires_at);
+	if (expires_at == NULL) {
+		return;
+	}
 
 	uint32_t curr_rdata = knot_rrsig_sig_expiration(rrsig);
 	knot_time_t current = knot_time_from_u32(curr_rdata);
 	*expires_at = knot_time_min(current, *expires_at);
 }
 
-/*!
- * \brief Add expired or invalid RRSIGs into the changeset for removal.
- *
- * \param covered     RR set with covered records.
- * \param rrsigs      RR set with RRSIGs.
- * \param sign_ctx    Local zone signing context.
- * \param changeset   Changeset to be updated.
- * \param expires_at  Earliest RRSIG expiration.
- *
- * \return Error code, KNOT_EOK if successful.
- */
-static int remove_expired_rrsigs(const knot_rrset_t *covered,
-                                 const knot_rrset_t *rrsigs,
-                                 zone_sign_ctx_t *sign_ctx,
-                                 changeset_t *changeset,
-                                 knot_time_t *expires_at)
+// TODO: move somewhere?
+static bool rrsig_covers_type(const knot_rrset_t *rrsig, uint16_t type)
 {
-	assert(changeset);
-
-	if (knot_rrset_empty(rrsigs)) {
-		return KNOT_EOK;
+	if (knot_rrset_empty(rrsig)) {
+		return false;
 	}
-
-	assert(rrsigs->type == KNOT_RRTYPE_RRSIG);
-
-	knot_rrset_t to_remove;
-	knot_rrset_init_empty(&to_remove);
-	int result = KNOT_EOK;
-
-	knot_rrset_t synth_rrsig = rrset_init_from(rrsigs, KNOT_RRTYPE_RRSIG);
-	result = knot_synth_rrsig(covered->type, &rrsigs->rrs, &synth_rrsig.rrs, NULL);
-	if (result != KNOT_EOK) {
-		if (result != KNOT_ENOENT) {
-			return result;
+	assert(rrsig->type == KNOT_RRTYPE_RRSIG);
+	knot_rdata_t *one_rr = rrsig->rrs.rdata;
+	for (int i = 0; i < rrsig->rrs.count; i++) {
+		if (type == knot_rrsig_type_covered(one_rr)) {
+			return true;
 		}
-		return KNOT_EOK;
+		one_rr = knot_rdataset_next(one_rr);
 	}
-
-	uint16_t rrsig_rdata_count = synth_rrsig.rrs.count;
-	for (uint16_t i = 0; i < rrsig_rdata_count; i++) {
-		knot_rdata_t *rr = knot_rdataset_at(&synth_rrsig.rrs, i);
-		uint16_t keytag = knot_rrsig_key_tag(rr);
-		int endloop = 0; // 1 - continue; 2 - break
-
-		for (size_t j = 0; j < sign_ctx->count; j++) {
-			zone_key_t *key = &sign_ctx->keys[j];
-
-			if (!key->is_active || dnssec_key_get_keytag(key->key) != keytag) {
-				continue;
-			}
-
-			result = knot_check_signature(covered, &synth_rrsig, i, key->key,
-			                              sign_ctx->sign_ctxs[j], sign_ctx->dnssec_ctx);
-			if (result == KNOT_EOK) {
-				// valid signature
-				note_earliest_expiration(rr, expires_at);
-				endloop = 1;
-				break;
-			} else if (result != DNSSEC_INVALID_SIGNATURE) {
-				endloop = 2;
-				break;
-			}
-		}
-
-		if (endloop == 2) {
-			break;
-		} else if (endloop == 1) {
-			continue;
-		}
-
-		if (knot_rrset_empty(&to_remove)) {
-			to_remove = create_empty_rrsigs_for(&synth_rrsig);
-		}
-
-		result = knot_rdataset_add(&to_remove.rrs, rr, NULL);
-		if (result != KNOT_EOK) {
-			break;
-		}
-	}
-
-	if (!knot_rrset_empty(&to_remove) && result == KNOT_EOK) {
-		result = changeset_add_removal(changeset, &to_remove, 0);
-	}
-
-	knot_rdataset_clear(&synth_rrsig.rrs, NULL);
-	knot_rdataset_clear(&to_remove.rrs, NULL);
-
-	return result;
+	return false;
 }
 
 /*!
  * \brief Add missing RRSIGs into the changeset for adding.
  *
+ * \note Also removes invalid RRSIGs.
+ *
  * \param covered     RR set with covered records.
  * \param rrsigs      RR set with RRSIGs.
  * \param sign_ctx    Local zone signing context.
+ * \param skip_crypto All RRSIGs in this node have been verified, just check validity.
  * \param changeset   Changeset to be updated.
+ * \param update      Zone update to be updated. Exactly one of "changeset" and "update" must be NULL!
  * \param expires_at  Earliest RRSIG expiration.
  *
  * \return Error code, KNOT_EOK if successful.
@@ -281,44 +225,72 @@ static int remove_expired_rrsigs(const knot_rrset_t *covered,
 static int add_missing_rrsigs(const knot_rrset_t *covered,
                               const knot_rrset_t *rrsigs,
                               zone_sign_ctx_t *sign_ctx,
+                              bool skip_crypto,
                               changeset_t *changeset,
+                              zone_update_t *update,
                               knot_time_t *expires_at)
 {
 	assert(!knot_rrset_empty(covered));
 	assert(sign_ctx);
-	assert(changeset);
+	assert((bool)changeset != (bool)update);
 
-	int result = KNOT_EOK;
-	knot_rrset_t to_add;
-	knot_rrset_init_empty(&to_add);
+	knot_rrset_t to_add = create_empty_rrsigs_for(covered);
+	knot_rrset_t to_remove = create_empty_rrsigs_for(covered);
+	int result = (!rrsig_covers_type(rrsigs, covered->type) ? KNOT_EOK :
+	             knot_synth_rrsig(covered->type, &rrsigs->rrs, &to_remove.rrs, NULL));
 
-	for (size_t i = 0; i < sign_ctx->count; i++) {
+	if (result == KNOT_EOK && sign_ctx->dnssec_ctx->offline_rrsig != NULL &&
+	    knot_dname_cmp(sign_ctx->dnssec_ctx->offline_rrsig->owner, covered->owner) == 0 &&
+	    rrsig_covers_type(sign_ctx->dnssec_ctx->offline_rrsig, covered->type)) {
+		result = knot_synth_rrsig(covered->type,
+		    &sign_ctx->dnssec_ctx->offline_rrsig->rrs, &to_add.rrs, NULL);
+		if (result == KNOT_EOK) {
+			// don't remove what shall be added
+			result = knot_rdataset_subtract(&to_remove.rrs, &to_add.rrs, NULL);
+		}
+		if (result == KNOT_EOK && !knot_rrset_empty(rrsigs)) {
+			// don't add what's already present
+			result = knot_rdataset_subtract(&to_add.rrs, &rrsigs->rrs, NULL);
+		}
+	}
+
+	for (size_t i = 0; i < sign_ctx->count && result == KNOT_EOK; i++) {
 		const zone_key_t *key = &sign_ctx->keys[i];
 		if (!knot_zone_sign_use_key(key, covered)) {
 			continue;
 		}
 
+		uint16_t valid_at;
 		if (valid_signature_exists(covered, rrsigs, key->key, sign_ctx->sign_ctxs[i],
-		                           sign_ctx->dnssec_ctx)) {
+		                           sign_ctx->dnssec_ctx, skip_crypto, &valid_at)) {
+			knot_rdata_t *valid_rr = knot_rdataset_at(&rrsigs->rrs, valid_at);
+			result = knot_rdataset_remove(&to_remove.rrs, valid_rr, NULL);
+			note_earliest_expiration(valid_rr, expires_at);
 			continue;
 		}
 
-		if (knot_rrset_empty(&to_add)) {
-			to_add = create_empty_rrsigs_for(covered);
-		}
-
 		result = knot_sign_rrset(&to_add, covered, key->key, sign_ctx->sign_ctxs[i],
 		                         sign_ctx->dnssec_ctx, NULL, expires_at);
-		if (result != KNOT_EOK) {
-			break;
+	}
+
+	if (!knot_rrset_empty(&to_remove) && result == KNOT_EOK) {
+		if (changeset != NULL) {
+			result = changeset_add_removal(changeset, &to_remove, 0);
+		} else {
+			result = zone_update_remove(update, &to_remove);
 		}
 	}
 
 	if (!knot_rrset_empty(&to_add) && result == KNOT_EOK) {
-		result = changeset_add_addition(changeset, &to_add, 0);
+		if (changeset != NULL) {
+			result = changeset_add_addition(changeset, &to_add, 0);
+		} else {
+			result = zone_update_add(update, &to_add);
+		}
 	}
 
 	knot_rdataset_clear(&to_add.rrs, NULL);
+	knot_rdataset_clear(&to_remove.rrs, NULL);
 
 	return result;
 }
@@ -379,7 +351,7 @@ static int force_resign_rrset(const knot_rrset_t *covered,
 		}
 	}
 
-	return add_missing_rrsigs(covered, NULL, sign_ctx, changeset, NULL);
+	return add_missing_rrsigs(covered, NULL, sign_ctx, false, changeset, NULL, NULL);
 }
 
 /*!
@@ -388,6 +360,7 @@ static int force_resign_rrset(const knot_rrset_t *covered,
  * \param covered     RR set with covered records.
  * \param rrsigs      Existing RRSIGs for covered RR set.
  * \param sign_ctx    Local zone signing context.
+ * \param skip_crypto All RRSIGs in this node have been verified, just check validity.
  * \param changeset   Changeset to be updated.
  * \param expires_at  Current earliest expiration, will be updated.
  *
@@ -396,19 +369,13 @@ static int force_resign_rrset(const knot_rrset_t *covered,
 static int resign_rrset(const knot_rrset_t *covered,
                         const knot_rrset_t *rrsigs,
                         zone_sign_ctx_t *sign_ctx,
+                        bool skip_crypto,
                         changeset_t *changeset,
                         knot_time_t *expires_at)
 {
 	assert(!knot_rrset_empty(covered));
 
-	// TODO this function creates some signatures twice (for checking)
-	int result = remove_expired_rrsigs(covered, rrsigs, sign_ctx,
-	                                   changeset, expires_at);
-	if (result != KNOT_EOK) {
-		return result;
-	}
-
-	return add_missing_rrsigs(covered, rrsigs, sign_ctx, changeset, expires_at);
+	return add_missing_rrsigs(covered, rrsigs, sign_ctx, skip_crypto, changeset, NULL, expires_at);
 }
 
 static int remove_standalone_rrsigs(const zone_node_t *node,
@@ -463,6 +430,8 @@ static int sign_node_rrsets(const zone_node_t *node,
 
 	int result = KNOT_EOK;
 	knot_rrset_t rrsigs = node_rrset(node, KNOT_RRTYPE_RRSIG);
+	bool skip_crypto = (node->flags & NODE_FLAGS_RRSIGS_VALID) &&
+	                   !sign_ctx->dnssec_ctx->keytag_conflict;
 
 	for (int i = 0; result == KNOT_EOK && i < node->rrset_count; i++) {
 		knot_rrset_t rrset = node_rrset_at(node, i);
@@ -475,7 +444,7 @@ static int sign_node_rrsets(const zone_node_t *node,
 			result = force_resign_rrset(&rrset, &rrsigs,
 			                            sign_ctx, changeset);
 		} else {
-			result = resign_rrset(&rrset, &rrsigs, sign_ctx,
+			result = resign_rrset(&rrset, &rrsigs, sign_ctx, skip_crypto,
 			                      changeset, expires_at);
 		}
 	}
@@ -489,10 +458,17 @@ static int sign_node_rrsets(const zone_node_t *node,
 /*!
  * \brief Struct to carry data for 'sign_data' callback function.
  */
-typedef struct node_sign_args {
+typedef struct {
+	zone_tree_t *tree;
 	zone_sign_ctx_t *sign_ctx;
-	changeset_t *changeset;
+	changeset_t changeset;
 	knot_time_t expires_at;
+	size_t num_threads;
+	size_t thread_index;
+	size_t rrset_index;
+	int errcode;
+	int thread_init_errcode;
+	pthread_t thread;
 } node_sign_args_t;
 
 /*!
@@ -501,59 +477,130 @@ typedef struct node_sign_args {
  * \param node  Node to be signed.
  * \param data  Callback data, node_sign_args_t.
  */
-static int sign_node(zone_node_t **node, void *data)
+static int sign_node(zone_node_t *node, void *data)
 {
-	assert(node && *node);
+	assert(node);
 	assert(data);
 
 	node_sign_args_t *args = (node_sign_args_t *)data;
 
-	if ((*node)->rrset_count == 0) {
+	if (node->rrset_count == 0) {
 		return KNOT_EOK;
 	}
 
-	int result = sign_node_rrsets(*node, args->sign_ctx, args->changeset,
-	                              &args->expires_at);
+	if (args->rrset_index++ % args->num_threads != args->thread_index) {
+		return KNOT_EOK;
+	}
+
+	int result = sign_node_rrsets(node, args->sign_ctx,
+	                              &args->changeset, &args->expires_at);
 
 	return result;
 }
 
+static void *tree_sign_thread(void *_arg)
+{
+	node_sign_args_t *arg = _arg;
+	arg->errcode = zone_tree_apply(arg->tree, sign_node, _arg);
+	return NULL;
+}
+
+static int set_signed(zone_node_t *node, void *data)
+{
+	UNUSED(data);
+	node->flags |= NODE_FLAGS_RRSIGS_VALID;
+	return KNOT_EOK;
+}
+
 /*!
  * \brief Update RRSIGs in a given zone tree by updating changeset.
  *
  * \param tree        Zone tree to be signed.
+ * \param num_threads Number of threads to use for parallel signing.
  * \param zone_keys   Zone keys.
  * \param policy      DNSSEC policy.
- * \param changeset   Changeset to be updated.
+ * \param update      Zone update structure to be updated.
  * \param expires_at  Expiration time of the oldest signature in zone.
  *
  * \return Error code, KNOT_EOK if successful.
  */
 static int zone_tree_sign(zone_tree_t *tree,
-                          const zone_keyset_t *zone_keys,
+                          size_t num_threads,
+                          zone_keyset_t *zone_keys,
                           const kdnssec_ctx_t *dnssec_ctx,
-                          changeset_t *changeset,
+                          zone_update_t *update,
                           knot_time_t *expires_at)
 {
 	assert(zone_keys);
 	assert(dnssec_ctx);
-	assert(changeset);
+	assert(update);
 
-	node_sign_args_t args = {
-		.sign_ctx = zone_sign_ctx(zone_keys, dnssec_ctx),
-		.changeset = changeset,
-		.expires_at = knot_time_add(dnssec_ctx->now, dnssec_ctx->policy->rrsig_lifetime),
-	};
+	int ret = KNOT_EOK;
+	node_sign_args_t args[num_threads];
+	memset(args, 0, sizeof(args));
+	*expires_at = knot_time_plus(dnssec_ctx->now, dnssec_ctx->policy->rrsig_lifetime);
 
-	if (args.sign_ctx == NULL) {
-		return KNOT_ENOMEM;
+	// init context structures
+	for (size_t i = 0; i < num_threads; i++) {
+		args[i].tree = tree;
+		args[i].sign_ctx = zone_sign_ctx(zone_keys, dnssec_ctx);
+		if (args[i].sign_ctx == NULL) {
+			ret = KNOT_ENOMEM;
+			break;
+		}
+		ret = changeset_init(&args[i].changeset, update->zone->name);
+		if (ret != KNOT_EOK) {
+			break;
+		}
+		args[i].expires_at = 0;
+		args[i].num_threads = num_threads;
+		args[i].thread_index = i;
+		args[i].rrset_index = 0;
+		args[i].errcode = KNOT_EOK;
+		args[i].thread_init_errcode = -1;
+	}
+	if (ret != KNOT_EOK) {
+		for (size_t i = 0; i < num_threads; i++) {
+			changeset_clear(&args[i].changeset);
+			zone_sign_ctx_free(args[i].sign_ctx);
+		}
+		return ret;
 	}
 
-	int result = zone_tree_apply(tree, sign_node, &args);
-	*expires_at = args.expires_at;
+	if (num_threads == 1) {
+		args[0].thread_init_errcode = 0;
+		tree_sign_thread(&args[0]);
+	} else {
+		// start working threads
+		for (size_t i = 0; i < num_threads; i++) {
+			args[i].thread_init_errcode =
+				pthread_create(&args[i].thread, NULL, tree_sign_thread, &args[i]);
+		}
 
-	zone_sign_ctx_free(args.sign_ctx);
-	return result;
+		// join those threads that have been really started
+		for (size_t i = 0; i < num_threads; i++) {
+			if (args[i].thread_init_errcode == 0) {
+				args[i].thread_init_errcode = pthread_join(args[i].thread, NULL);
+			}
+		}
+	}
+
+	// collect return code and results
+	for (size_t i = 0; i < num_threads && ret == KNOT_EOK; i++) {
+		if (args[i].thread_init_errcode != 0) {
+			ret = knot_map_errno_code(args[i].thread_init_errcode);
+		} else {
+			ret = args[i].errcode;
+			if (ret == KNOT_EOK) {
+				ret = zone_update_apply_changeset(update, &args[i].changeset); // _fix not needed
+				*expires_at = knot_time_min(*expires_at, args[i].expires_at);
+			}
+		}
+		changeset_clear(&args[i].changeset);
+		zone_sign_ctx_free(args[i].sign_ctx);
+	}
+
+	return ret;
 }
 
 /*- private API - signing of NSEC(3) in changeset ----------------------------*/
@@ -563,17 +610,23 @@ static int zone_tree_sign(zone_tree_t *tree,
  */
 typedef struct {
 	const zone_contents_t *zone;
+	changeset_iter_t itt;
 	zone_sign_ctx_t *sign_ctx;
-	changeset_t *changeset;
-	trie_t *signed_tree;
+	changeset_t changeset;
+	knot_time_t expires_at;
+	size_t num_threads;
+	size_t thread_index;
+	size_t rrset_index;
+	int errcode;
+	int thread_init_errcode;
+	pthread_t thread;
 } changeset_signing_data_t;
 
-/*- private API - DNSKEY handling --------------------------------------------*/
-
-static int rrset_add_zone_key(knot_rrset_t *rrset, zone_key_t *zone_key)
+int rrset_add_zone_key(knot_rrset_t *rrset, zone_key_t *zone_key)
 {
-	assert(rrset);
-	assert(zone_key);
+	if (rrset == NULL || zone_key == NULL) {
+		return KNOT_EINVAL;
+	}
 
 	dnssec_binary_t dnskey_rdata = { 0 };
 	dnssec_key_get_rdata(zone_key->key, &dnskey_rdata);
@@ -592,256 +645,121 @@ static int rrset_add_zone_ds(knot_rrset_t *rrset, zone_key_t *zone_key)
 	return knot_rrset_add_rdata(rrset, cds_rdata.data, cds_rdata.size, NULL);
 }
 
-/*!
- * \brief Goes through list and looks for RRSet type there.
- *
- * \return True if RR type is in the list, false otherwise.
- */
-static bool rr_type_in_list(const knot_rrset_t *rr, const list_t *l)
-{
-	if (l == NULL || EMPTY_LIST(*l)) {
-		return false;
-	}
-	assert(rr);
-
-	type_node_t *n = NULL;
-	WALK_LIST(n, *l) {
-		type_node_t *type_node = (type_node_t *)n;
-		if (type_node->type == rr->type) {
-			return true;
-		}
-	};
-
-	return false;
-}
-
-static int add_rr_type_to_list(const knot_rrset_t *rr, list_t *l)
-{
-	assert(rr);
-	assert(l);
-
-	type_node_t *n = malloc(sizeof(type_node_t));
-	if (n == NULL) {
-		return KNOT_ENOMEM;
-	}
-	n->type = rr->type;
-
-	add_head(l, (node_t *)n);
-	return KNOT_EOK;
-}
-
-/*!
- * \brief Checks whether RRSet is not already in the hash table, automatically
- *        stores its pointer to the table if not found, but returns false in
- *        that case.
- *
- * \param rrset      RRSet to be checked for.
- * \param tree       Tree with already signed RRs.
- * \param rr_signed  Set to true if RR is signed already, false otherwise.
- *
- * \return KNOT_E*
- */
-static int rr_already_signed(const knot_rrset_t *rrset, trie_t *t,
-                             bool *rr_signed)
-{
-	assert(rrset);
-	assert(t);
-	*rr_signed = false;
-	// Create a key = RRSet owner converted to sortable format
-	knot_dname_storage_t lf_storage;
-	uint8_t *lf = knot_dname_lf(rrset->owner, lf_storage);
-	assert(lf);
-	trie_val_t stored_info = (signed_info_t *)trie_get_try(t, (char *)lf+1,
-	                                                      *lf);
-	if (stored_info == NULL) {
-		// Create new info struct
-		signed_info_t *info = malloc(sizeof(signed_info_t));
-		if (info == NULL) {
-			return KNOT_ENOMEM;
-		}
-		memset(info, 0, sizeof(signed_info_t));
-		// Store actual dname repr
-		info->dname = knot_dname_copy(rrset->owner, NULL);
-		if (info->dname == NULL) {
-			free(info);
-			return KNOT_ENOMEM;
-		}
-		// Create new list to insert as a value
-		info->type_list = malloc(sizeof(list_t));
-		if (info->type_list == NULL) {
-			free(info->dname);
-			free(info);
-			return KNOT_ENOMEM;
-		}
-		init_list(info->type_list);
-		// Insert type to list
-		int ret = add_rr_type_to_list(rrset, info->type_list);
-		if (ret != KNOT_EOK) {
-			free(info->type_list);
-			free(info->dname);
-			free(info);
-			return ret;
-		}
-		*trie_get_ins(t, (char *)lf+1, *lf) = info;
-	} else {
-		signed_info_t *info = *((signed_info_t **)stored_info);
-		assert(info->type_list);
-		// Check whether the type is in the list already
-		if (rr_type_in_list(rrset, info->type_list)) {
-			*rr_signed = true;
-			return KNOT_EOK;
-		}
-		// Just update the existing list
-		int ret = add_rr_type_to_list(rrset, info->type_list);
-		if (ret != KNOT_EOK) {
-			*rr_signed = false;
-			return KNOT_EOK;
-		}
-	}
-
-	*rr_signed = false;
-	return KNOT_EOK;
-}
-
-/*!
- * \brief Wrapper function for changeset signing - to be used with changeset
- *        apply functions.
- *
- * \param chg_rrset  RRSet to be signed (potentially)
- * \param data       Signing data
- *
- * \return Error code, KNOT_EOK if successful.
- */
-static int sign_changeset_wrap(knot_rrset_t *chg_rrset,
-                               changeset_signing_data_t *args,
-                               knot_time_t *expire_at)
-{
-	// Find RR's node in zone, find out if we need to sign this RR
-	const zone_node_t *node =
-		zone_contents_find_node(args->zone, chg_rrset->owner);
-
-	// If node is not in zone, all its RRSIGs were dropped - no-op
-	if (node) {
-		knot_rrset_t zone_rrset = node_rrset(node, chg_rrset->type);
-		knot_rrset_t rrsigs = node_rrset(node, KNOT_RRTYPE_RRSIG);
-
-		bool should_sign = knot_zone_sign_rr_should_be_signed(node, &zone_rrset);
-
-		// Check for RRSet in the 'already_signed' table
-		if (args->signed_tree && (should_sign && knot_rrset_empty(&zone_rrset))) {
-			bool already_signed = false;
-
-			int ret = rr_already_signed(chg_rrset, args->signed_tree,
-			                            &already_signed);
-			if (ret != KNOT_EOK) {
-				return ret;
-			}
-			if (already_signed) {
-				/* Do not sign again. */
-				should_sign = false;
-			}
-		}
-
-		if (should_sign) {
-			return resign_rrset(&zone_rrset, &rrsigs, args->sign_ctx,
-			                    args->changeset, expire_at);
-		} else {
-			/*
-			 * If RRSet in zone DOES have RRSIGs although we
-			 * should not sign it, DDNS-caused change to node/rr
-			 * occurred and we have to drop all RRSIGs.
-			 *
-			 * OR
-			 *
-			 * The whole RRSet was removed, but RRSIGs remained in
-			 * the zone. We need to drop them as well.
-			 */
-			return remove_rrset_rrsigs(chg_rrset->owner,
-			                           chg_rrset->type, &rrsigs,
-			                           args->changeset);
-		}
-	}
-
-	return KNOT_EOK;
-}
-
-/*!
- * \brief Frees info node about update signing.
- *
- * \param val  Node to free.
- * \param d    Unused.
- */
-static int free_helper_trie_node(trie_val_t *val, void *d)
-{
-	UNUSED(d);
-	signed_info_t *info = (signed_info_t *)*val;
-	if (info->type_list && !EMPTY_LIST(*(info->type_list))) {
-		WALK_LIST_FREE(*(info->type_list));
-	}
-	free(info->type_list);
-	knot_dname_free(info->dname, NULL);
-	knot_dname_free(info->hashed_dname, NULL);
-	free(info);
-	return KNOT_EOK;
-}
-
-/*!
- * \brief Clears trie with info about update signing.
- *
- * \param t  Trie to clear.
- */
-static void knot_zone_clear_sorted_changes(trie_t *t)
-{
-	if (t) {
-		trie_apply(t, free_helper_trie_node, NULL);
-	}
-}
-
-/*- public API ---------------------------------------------------------------*/
-
 int knot_zone_sign(zone_update_t *update,
                    zone_keyset_t *zone_keys,
                    const kdnssec_ctx_t *dnssec_ctx,
                    knot_time_t *expire_at)
 {
-	if (!update || !zone_keys || !dnssec_ctx || !expire_at) {
+	if (!update || !zone_keys || !dnssec_ctx || !expire_at ||
+	    dnssec_ctx->policy->signing_threads < 1) {
 		return KNOT_EINVAL;
 	}
 
 	int result;
 
-	changeset_t ch;
-	result = changeset_init(&ch, update->new_cont->apex->owner);
-	if (result != KNOT_EOK) {
-		return result;
-	}
-
 	knot_time_t normal_expire = 0;
-	result = zone_tree_sign(update->new_cont->nodes, zone_keys, dnssec_ctx, &ch, &normal_expire);
+	result = zone_tree_sign(update->new_cont->nodes, dnssec_ctx->policy->signing_threads,
+	                        zone_keys, dnssec_ctx, update, &normal_expire);
 	if (result != KNOT_EOK) {
-		changeset_clear(&ch);
 		return result;
 	}
 
 	knot_time_t nsec3_expire = 0;
-	result = zone_tree_sign(update->new_cont->nsec3_nodes, zone_keys, dnssec_ctx,
-				&ch, &nsec3_expire);
+	result = zone_tree_sign(update->new_cont->nsec3_nodes, dnssec_ctx->policy->signing_threads,
+	                        zone_keys, dnssec_ctx, update, &nsec3_expire);
 	if (result != KNOT_EOK) {
-		changeset_clear(&ch);
 		return result;
 	}
 
+	bool whole = !(update->flags & UPDATE_INCREMENTAL);
+	result = zone_tree_apply(whole ? update->new_cont->nodes : update->a_ctx->node_ptrs, set_signed, NULL);
+	if (result == KNOT_EOK) {
+		result = zone_tree_apply(whole ? update->new_cont->nsec3_nodes : update->a_ctx->nsec3_ptrs, set_signed, NULL);
+	}
+
 	*expire_at = knot_time_min(normal_expire, nsec3_expire);
 
-	result = zone_update_apply_changeset(update, &ch); // _fix not needed
-	changeset_clear(&ch);
-
 	return result;
 }
 
+keyptr_dynarray_t knot_zone_sign_get_cdnskeys(const kdnssec_ctx_t *ctx,
+					      zone_keyset_t *zone_keys)
+{
+	keyptr_dynarray_t r = { 0 };
+	unsigned crp = ctx->policy->cds_cdnskey_publish;
+
+	if (crp == CDS_CDNSKEY_ROLLOVER || crp == CDS_CDNSKEY_ALWAYS ||
+	    crp == CDS_CDNSKEY_DOUBLE_DS) {
+		// first, add strictly-ready keys
+		for (int i = 0; i < zone_keys->count; i++) {
+			zone_key_t *key = &zone_keys->keys[i];
+			if (key->is_ready) {
+				assert(key->is_ksk);
+				keyptr_dynarray_add(&r, &key);
+			}
+		}
+
+		// second, add active keys
+		if ((crp == CDS_CDNSKEY_ALWAYS && r.size == 0) ||
+		    (crp == CDS_CDNSKEY_DOUBLE_DS)) {
+			for (int i = 0; i < zone_keys->count; i++) {
+				zone_key_t *key = &zone_keys->keys[i];
+				if (key->is_ksk && key->is_active && !key->is_ready) {
+					keyptr_dynarray_add(&r, &key);
+				}
+			}
+		}
+
+		if ((crp != CDS_CDNSKEY_DOUBLE_DS && r.size > 1) ||
+		    (r.size > 2)) {
+			log_zone_warning(ctx->zone->dname, "DNSSEC, published CDS/CDNSKEY records for too many (%zu) keys", r.size);
+		}
+	}
+
+	return r;
+}
+
+int knot_zone_sign_add_dnskeys(zone_keyset_t *zone_keys, const kdnssec_ctx_t *dnssec_ctx,
+			       key_records_t *add_r)
+{
+	if (add_r == NULL) {
+		return KNOT_EINVAL;
+	}
+	int ret = KNOT_EOK;
+	for (int i = 0; i < zone_keys->count; i++) {
+		zone_key_t *key = &zone_keys->keys[i];
+		if (key->is_public) {
+			ret = rrset_add_zone_key(&add_r->dnskey, key);
+			if (ret != KNOT_EOK) {
+				return ret;
+			}
+		}
+	}
+	keyptr_dynarray_t kcdnskeys = knot_zone_sign_get_cdnskeys(dnssec_ctx, zone_keys);
+	dynarray_foreach(keyptr, zone_key_t *, ksk_for_cds, kcdnskeys) {
+		ret = rrset_add_zone_key(&add_r->cdnskey, *ksk_for_cds);
+		if (ret == KNOT_EOK) {
+			ret = rrset_add_zone_ds(&add_r->cds, *ksk_for_cds);
+		}
+	}
+
+	if (dnssec_ctx->policy->cds_cdnskey_publish == CDS_CDNSKEY_EMPTY && ret == KNOT_EOK) {
+		const uint8_t cdnskey_empty[5] = { 0, 0, 3, 0, 0 };
+		const uint8_t cds_empty[5] = { 0, 0, 0, 0, 0 };
+		ret = knot_rrset_add_rdata(&add_r->cdnskey, cdnskey_empty, sizeof(cdnskey_empty), NULL);
+		if (ret == KNOT_EOK) {
+			ret = knot_rrset_add_rdata(&add_r->cds, cds_empty, sizeof(cds_empty), NULL);
+		}
+	}
+
+	keyptr_dynarray_free(&kcdnskeys);
+	return ret;
+}
+
 int knot_zone_sign_update_dnskeys(zone_update_t *update,
                                   zone_keyset_t *zone_keys,
-                                  const kdnssec_ctx_t *dnssec_ctx)
+                                  kdnssec_ctx_t *dnssec_ctx,
+                                  knot_time_t *next_resign)
 {
 	if (update == NULL || zone_keys == NULL || dnssec_ctx == NULL) {
 		return KNOT_EINVAL;
@@ -851,10 +769,8 @@ int knot_zone_sign_update_dnskeys(zone_update_t *update,
 	knot_rrset_t dnskeys = node_rrset(apex, KNOT_RRTYPE_DNSKEY);
 	knot_rrset_t cdnskeys = node_rrset(apex, KNOT_RRTYPE_CDNSKEY);
 	knot_rrset_t cdss = node_rrset(apex, KNOT_RRTYPE_CDS);
-	knot_rrset_t *add_dnskeys = NULL;
-	knot_rrset_t *add_cdnskeys = NULL;
-	knot_rrset_t *add_cdss = NULL;
-	uint32_t dnskey_ttl = dnssec_ctx->policy->dnskey_ttl;
+	key_records_t add_r;
+	memset(&add_r, 0, sizeof(add_r));
 	knot_rrset_t soa = node_rrset(apex, KNOT_RRTYPE_SOA);
 	if (knot_rrset_empty(&soa)) {
 		return KNOT_EINVAL;
@@ -877,67 +793,45 @@ int knot_zone_sign_update_dnskeys(zone_update_t *update,
 	CHECK_RET;
 
 	// add DNSKEYs, CDNSKEYs and CDSs
-	add_dnskeys = knot_rrset_new(apex->owner, KNOT_RRTYPE_DNSKEY, soa.rclass,
-	                             dnskey_ttl, NULL);
-	add_cdnskeys = knot_rrset_new(apex->owner, KNOT_RRTYPE_CDNSKEY, soa.rclass,
-	                              0, NULL);
-	add_cdss = knot_rrset_new(apex->owner, KNOT_RRTYPE_CDS, soa.rclass,
-	                          0, NULL);
-	if (add_dnskeys == NULL || add_cdnskeys == NULL || add_cdss == NULL) {
-		ret = KNOT_ENOMEM;
-		CHECK_RET;
-	}
-	zone_key_t *ksk_for_cds = NULL;
-	unsigned crp = dnssec_ctx->policy->child_records_publish;
-	int kfc_prio = (crp == CHILD_RECORDS_ALWAYS ? 0 : (crp == CHILD_RECORDS_ROLLOVER ? 1 : 2));
-	for (int i = 0; i < zone_keys->count; i++) {
-		zone_key_t *key = &zone_keys->keys[i];
-		if (key->is_public) {
-			ret = rrset_add_zone_key(add_dnskeys, key);
-			CHECK_RET;
+	key_records_init(dnssec_ctx, &add_r);
+
+	if (dnssec_ctx->policy->offline_ksk) {
+		ret = kasp_db_load_offline_records(dnssec_ctx->kasp_db, apex->owner, dnssec_ctx->now, next_resign, &add_r);
+		if (ret == KNOT_EOK) {
+			log_zone_info(dnssec_ctx->zone->dname,
+			              "DNSSEC, using offline records, DNSKEYs %hu, CDNSKEYs %hu, CDs %hu, RRSIGs %hu",
+			              add_r.dnskey.rrs.count, add_r.cdnskey.rrs.count, add_r.cds.rrs.count, add_r.rrsig.rrs.count);
+		} else {
+			log_zone_warning(dnssec_ctx->zone->dname, "DNSSEC, failed to load offline records (%s)",
+			                 knot_strerror(ret));
 		}
+	} else {
+		ret = knot_zone_sign_add_dnskeys(zone_keys, dnssec_ctx, &add_r);
+	}
+	CHECK_RET;
 
-		// determine which key (if any) will be the one for CDS/CDNSKEY
-		if (key->is_ksk && key->cds_priority > kfc_prio) {
-			ksk_for_cds = key;
-			kfc_prio = key->cds_priority;
+	if (!knot_rrset_empty(&add_r.cdnskey)) {
+		ret = changeset_add_addition(&ch, &add_r.cdnskey, CHANGESET_CHECK);
+		CHECK_RET;
+	}
+
+	if (!knot_rrset_empty(&add_r.cds)) {
+		ret = changeset_add_addition(&ch, &add_r.cds, CHANGESET_CHECK);
+		if (node_rrtype_exists(ch.add->apex, KNOT_RRTYPE_CDS)) {
+			// there is indeed a change to CDS
+			update->zone->timers.next_ds_push = time(NULL);
+			zone_events_schedule_now(update->zone, ZONE_EVENT_DS_PUSH);
 		}
-	}
-
-	if (ksk_for_cds != NULL) {
-		ret = rrset_add_zone_key(add_cdnskeys, ksk_for_cds);
-		CHECK_RET;
-		ret = rrset_add_zone_ds(add_cdss, ksk_for_cds);
 		CHECK_RET;
 	}
 
-	if (crp == CHILD_RECORDS_EMPTY) {
-		const uint8_t cdnskey_empty[5] = { 0, 0, 3, 0, 0 };
-		const uint8_t cds_empty[5] = { 0, 0, 0, 0, 0 };
-		ret = knot_rrset_add_rdata(add_cdnskeys, cdnskey_empty,
-		                           sizeof(cdnskey_empty), NULL);
-		CHECK_RET;
-		ret = knot_rrset_add_rdata(add_cdss, cds_empty,
-		                           sizeof(cds_empty), NULL);
+	if (!knot_rrset_empty(&add_r.dnskey)) {
+		ret = changeset_add_addition(&ch, &add_r.dnskey, CHANGESET_CHECK);
 		CHECK_RET;
 	}
 
-	if (!knot_rrset_empty(add_cdnskeys)) {
-		ret = changeset_add_addition(&ch, add_cdnskeys, CHANGESET_CHECK |
-		                                                CHANGESET_CHECK_CANCELOUT);
-		CHECK_RET;
-	}
-
-	if (!knot_rrset_empty(add_cdss)) {
-		ret = changeset_add_addition(&ch, add_cdss, CHANGESET_CHECK |
-		                                            CHANGESET_CHECK_CANCELOUT);
-		CHECK_RET;
-	}
-
-	if (!knot_rrset_empty(add_dnskeys)) {
-		ret = changeset_add_addition(&ch, add_dnskeys, CHANGESET_CHECK |
-		                                               CHANGESET_CHECK_CANCELOUT);
-		CHECK_RET;
+	if (!knot_rrset_empty(&add_r.rrsig)) {
+		dnssec_ctx->offline_rrsig = knot_rrset_copy(&add_r.rrsig, NULL);
 	}
 
 	ret = zone_update_apply_changeset(update, &ch);
@@ -945,9 +839,7 @@ int knot_zone_sign_update_dnskeys(zone_update_t *update,
 #undef CHECK_RET
 
 cleanup:
-	knot_rrset_free(add_dnskeys, NULL);
-	knot_rrset_free(add_cdnskeys, NULL);
-	knot_rrset_free(add_cdss, NULL);
+	key_records_clear(&add_r);
 	changeset_clear(&ch);
 	return ret;
 }
@@ -958,9 +850,8 @@ bool knot_zone_sign_use_key(const zone_key_t *key, const knot_rrset_t *covered)
 		return false;
 	}
 
-	if (!key->is_active) {
-		return false;
-	}
+	bool active_ksk = ((key->is_active || key->is_ksk_active_plus) && key->is_ksk);
+	bool active_zsk = ((key->is_active || key->is_zsk_active_plus) && key->is_zsk);;
 
 	// this may be a problem with offline KSK
 	bool cds_sign_by_ksk = true;
@@ -969,17 +860,17 @@ bool knot_zone_sign_use_key(const zone_key_t *key, const knot_rrset_t *covered)
 	bool is_apex = knot_dname_is_equal(covered->owner,
 	                                   dnssec_key_get_dname(key->key));
 	if (!is_apex) {
-		return key->is_zsk;
+		return active_zsk;
 	}
 
 	switch (covered->type) {
 	case KNOT_RRTYPE_DNSKEY:
-		return key->is_ksk;
+		return active_ksk;
 	case KNOT_RRTYPE_CDS:
 	case KNOT_RRTYPE_CDNSKEY:
-		return (cds_sign_by_ksk ? key->is_ksk : key->is_zsk);
+		return (cds_sign_by_ksk ? active_ksk : active_zsk);
 	default:
-		return key->is_zsk;
+		return active_zsk;
 	}
 }
 
@@ -1003,72 +894,11 @@ bool knot_zone_sign_soa_expired(const zone_contents_t *zone,
 	return !exist;
 }
 
-static int sign_changeset(const zone_contents_t *zone,
-                          const changeset_t *in_ch,
-                          changeset_t *out_ch,
-                          const zone_keyset_t *zone_keys,
-                          const kdnssec_ctx_t *dnssec_ctx,
-                          knot_time_t *expire_at)
-{
-	if (zone == NULL || in_ch == NULL || out_ch == NULL) {
-		return KNOT_EINVAL;
-	}
-
-	int ret = KNOT_EOK;
-
-	// Create args for wrapper function - trie for duplicate sigs
-	changeset_signing_data_t args = {
-		.zone = zone,
-		.sign_ctx = zone_sign_ctx(zone_keys, dnssec_ctx),
-		.changeset = out_ch,
-		.signed_tree = trie_create(NULL)
-	};
-
-	if (args.sign_ctx == NULL || args.signed_tree == NULL) {
-		ret = KNOT_ENOMEM;
-		goto cleanup;
-	}
-
-	changeset_iter_t itt;
-	changeset_iter_all(&itt, in_ch);
-
-	knot_rrset_t rr = changeset_iter_next(&itt);
-	while (!knot_rrset_empty(&rr)) {
-		ret = sign_changeset_wrap(&rr, &args, expire_at);
-		if (ret != KNOT_EOK) {
-			changeset_iter_clear(&itt);
-			goto cleanup;
-		}
-		rr = changeset_iter_next(&itt);
-	}
-	changeset_iter_clear(&itt);
-
-	if (!knot_rrset_empty(in_ch->soa_from)) {
-		ret = sign_changeset_wrap(in_ch->soa_from, &args, expire_at);
-		if (ret != KNOT_EOK) {
-			goto cleanup;
-		}
-	}
-	if (!knot_rrset_empty(in_ch->soa_to)) {
-		ret = sign_changeset_wrap(in_ch->soa_to, &args, expire_at);
-		if (ret != KNOT_EOK) {
-			goto cleanup;
-		}
-	}
-
-cleanup:
-	knot_zone_clear_sorted_changes(args.signed_tree);
-	trie_free(args.signed_tree);
-	zone_sign_ctx_free(args.sign_ctx);
-
-	return ret;
-}
-
 int knot_zone_sign_nsecs_in_changeset(const zone_keyset_t *zone_keys,
                                       const kdnssec_ctx_t *dnssec_ctx,
-                                      changeset_t *changeset)
+                                      zone_update_t *update)
 {
-	if (zone_keys == NULL || dnssec_ctx == NULL || changeset == NULL) {
+	if (zone_keys == NULL || dnssec_ctx == NULL || update == NULL) {
 		return KNOT_EINVAL;
 	}
 
@@ -1077,28 +907,33 @@ int knot_zone_sign_nsecs_in_changeset(const zone_keyset_t *zone_keys,
 		return KNOT_ENOMEM;
 	}
 
-	changeset_iter_t itt;
-	changeset_iter_add(&itt, changeset);
+	zone_tree_it_t it = { 0 };
+	int ret = zone_tree_it_double_begin(update->a_ctx->node_ptrs, update->a_ctx->nsec3_ptrs, &it);
 
-	knot_rrset_t rr = changeset_iter_next(&itt);
-	while (!knot_rrset_empty(&rr)) {
-		if (rr.type == KNOT_RRTYPE_NSEC ||
-		    rr.type == KNOT_RRTYPE_NSEC3 ||
-		    rr.type == KNOT_RRTYPE_NSEC3PARAM) {
-			int ret =  add_missing_rrsigs(&rr, NULL, sign_ctx,
-			                              changeset, NULL);
-			if (ret != KNOT_EOK) {
-				changeset_iter_clear(&itt);
-				return ret;
+	while (!zone_tree_it_finished(&it) && ret == KNOT_EOK) {
+		zone_node_t *n = zone_tree_it_val(&it);
+		bool skip_crypto = (n->flags & NODE_FLAGS_RRSIGS_VALID) && !dnssec_ctx->keytag_conflict;
+
+		knot_rrset_t rrsigs = node_rrset(n, KNOT_RRTYPE_RRSIG);
+		for (int i = 0; i < n->rrset_count; i++) {
+			knot_rrset_t rr = node_rrset_at(n, i);
+			if (rr.type == KNOT_RRTYPE_NSEC ||
+			    rr.type == KNOT_RRTYPE_NSEC3 ||
+			    rr.type == KNOT_RRTYPE_NSEC3PARAM) {
+				ret =  add_missing_rrsigs(&rr, &rrsigs, sign_ctx, skip_crypto, NULL, update, NULL);
 			}
 		}
-		rr = changeset_iter_next(&itt);
-	}
 
-	changeset_iter_clear(&itt);
+		if (ret == KNOT_EOK) {
+			n->flags |= NODE_FLAGS_RRSIGS_VALID; // non-NSEC RRSIGs had been validated in knot_dnssec_sign_update()
+		}
+
+		zone_tree_it_next(&it);
+	}
+	zone_tree_it_free(&it);
 	zone_sign_ctx_free(sign_ctx);
 
-	return KNOT_EOK;
+	return ret;
 }
 
 bool knot_zone_sign_rr_should_be_signed(const zone_node_t *node,
@@ -1128,35 +963,24 @@ int knot_zone_sign_update(zone_update_t *update,
                           const kdnssec_ctx_t *dnssec_ctx,
                           knot_time_t *expire_at)
 {
-	if (update == NULL || zone_keys == NULL || dnssec_ctx == NULL || expire_at == NULL) {
+	if (update == NULL || zone_keys == NULL || dnssec_ctx == NULL || expire_at == NULL ||
+	    dnssec_ctx->policy->signing_threads < 1) {
 		return KNOT_EINVAL;
 	}
 
 	int ret = KNOT_EOK;
 
-
-	ret = apply_prepare_to_sign(update->a_ctx);
-	if (ret != KNOT_EOK) {
-		return ret;
-	}
-
 	/* Check if the UPDATE changed DNSKEYs or NSEC3PARAM.
 	 * If so, we have to sign the whole zone. */
 	const bool full_sign = apex_dnssec_changed(update);
 	if (full_sign) {
 		ret = knot_zone_sign(update, zone_keys, dnssec_ctx, expire_at);
 	} else {
-		changeset_t sec_ch;
-		ret = changeset_init(&sec_ch, update->zone->name);
-		if (ret != KNOT_EOK) {
-			return ret;
-		}
-		ret = sign_changeset(update->new_cont, &update->change, &sec_ch,
-		                     zone_keys, dnssec_ctx, expire_at);
+		ret = zone_tree_sign(update->a_ctx->node_ptrs, dnssec_ctx->policy->signing_threads,
+				     zone_keys, dnssec_ctx, update, expire_at);
 		if (ret == KNOT_EOK) {
-			ret = zone_update_apply_changeset_fix(update, &sec_ch);
+			ret = zone_tree_apply(update->a_ctx->node_ptrs, set_signed, NULL);
 		}
-		changeset_clear(&sec_ch);
 	}
 
 	return ret;
@@ -1178,7 +1002,7 @@ int knot_zone_sign_soa(zone_update_t *update,
 		}
 		ret = force_resign_rrset(&soa_to, &soa_rrsig, sign_ctx, &ch);
 		if (ret == KNOT_EOK) {
-			ret = zone_update_apply_changeset_fix(update, &ch);
+			ret = zone_update_apply_changeset(update, &ch);
 		}
 		zone_sign_ctx_free(sign_ctx);
 	}
