@@ -1,4 +1,4 @@
-/*  Copyright (C) 2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,13 +11,14 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #pragma once
 
 #include "libknot/packet/pkt.h"
 #include "knot/include/module.h"
+#include "knot/nameserver/process_query.h"
 
 /*!
  * \brief Answer query from an IN class zone.
@@ -50,24 +51,18 @@ int internet_process_query(knot_pkt_t *pkt, knotd_qdata_t *qdata);
 
 /*! \brief Require existing zone contents or return failure. */
 #define NS_NEED_ZONE_CONTENTS(qdata, error_rcode) \
-	if ((qdata)->extra->zone->contents == NULL) { \
+	if ((qdata)->extra->contents == NULL) { \
 		qdata->rcode = (error_rcode); \
 		return KNOT_STATE_FAIL; \
 	}
 
 /*! \brief Require authentication. */
-#define NS_NEED_AUTH(qdata, zone_name, action) \
-	if (!process_query_acl_check(conf(), (zone_name), (action), (qdata)) || \
+#define NS_NEED_AUTH(qdata, action) \
+	if (!process_query_acl_check(conf(), (action), (qdata)) || \
 	    process_query_verify(qdata) != KNOT_EOK) { \
 		return KNOT_STATE_FAIL; \
 	}
 
-/*! \brief Require maximum number of unsigned messages. */
-#define NS_NEED_TSIG_SIGNED(tsig_ctx, max_unsigned) \
-	if (tsig_unsigned_count(tsig_ctx) > max_unsigned) { \
-		return KNOT_STATE_FAIL; \
-	}
-
 /*! \brief Require the zone not to be frozen. */
 #define NS_NEED_NOT_FROZEN(qdata, error_rcode) \
 	if ((qdata)->extra->zone->events.ufrozen) { \
