@@ -1,4 +1,4 @@
-/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2019 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,14 +11,14 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-*/
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ */
 
 #include <urcu.h>
 
 #include "contrib/mempattern.h"
 #include "contrib/sockaddr.h"
-#include "knot/journal/chgset_ctx.h"
+#include "knot/journal/journal_metadata.h"
 #include "knot/nameserver/axfr.h"
 #include "knot/nameserver/internet.h"
 #include "knot/nameserver/ixfr.h"
@@ -43,77 +43,88 @@
 
 /*! \brief Puts current RR into packet, stores state for retries. */
 static int ixfr_put_chg_part(knot_pkt_t *pkt, struct ixfr_proc *ixfr,
-                             chgset_ctx_t *itt)
+                             journal_read_t *read)
 {
 	assert(pkt);
 	assert(ixfr);
-	assert(itt);
+	assert(read);
 
 	if (!knot_rrset_empty(&ixfr->cur_rr)) {
 		IXFR_SAFE_PUT(pkt, &ixfr->cur_rr);
-		knot_rrset_clear(&ixfr->cur_rr, NULL);
+		if (ixfr->cur_rr.type == KNOT_RRTYPE_SOA) {
+			ixfr->in_remove_section = !ixfr->in_remove_section;
+		}
+		journal_read_clear_rrset(&ixfr->cur_rr);
 	}
 
-	while (itt->phase != CHGSET_CTX_DONE) {
-		int r = chgset_ctx_next(itt, &ixfr->cur_rr);
-		if (r != KNOT_EOK) {
-			knot_rrset_clear(&ixfr->cur_rr, NULL);
-			return r;
+	while (journal_read_rrset(read, &ixfr->cur_rr, true)) {
+		if (ixfr->cur_rr.type == KNOT_RRTYPE_SOA &&
+		    !ixfr->in_remove_section &&
+		    knot_soa_serial(ixfr->cur_rr.rrs.rdata) == ixfr->soa_to) {
+			break;
 		}
 		IXFR_SAFE_PUT(pkt, &ixfr->cur_rr);
+		if (ixfr->cur_rr.type == KNOT_RRTYPE_SOA) {
+			ixfr->in_remove_section = !ixfr->in_remove_section;
+		}
 		knot_rrset_clear(&ixfr->cur_rr, NULL);
 	}
 
-	return KNOT_EOK;
+	return journal_read_get_error(read, KNOT_EOK);
 }
 
 /*!
- * \brief Process single changeset.
+ * \brief Process the changes from journal.
  * \note Keep in mind that this function must be able to resume processing,
  *       for example if it fills a packet and returns ESPACE, it is called again
  *       with next empty answer and it must resume the processing exactly where
  *       it's left off.
  */
-static int ixfr_process_changeset(knot_pkt_t *pkt, const void *item,
-                                  struct xfr_proc *xfer)
+static int ixfr_process_journal(knot_pkt_t *pkt, const void *item,
+                                struct xfr_proc *xfer)
 {
 	int ret = KNOT_EOK;
 	struct ixfr_proc *ixfr = (struct ixfr_proc *)xfer;
-	chgset_ctx_t *chgset = (chgset_ctx_t *)item;
+	journal_read_t *read = (journal_read_t *)item;
 
-	ret = ixfr_put_chg_part(pkt, ixfr, chgset);
-	if (ret != KNOT_EOK) {
-		return ret;
-	}
-
-	/* Finished change set. */
-	IXFROUT_LOG(LOG_DEBUG, ixfr->qdata, "serial %u -> %u", chgset->serial_from, chgset->serial_to);
+	ret = ixfr_put_chg_part(pkt, ixfr, read);
 
 	return ret;
 }
 
 #undef IXFR_SAFE_PUT
 
-static int ixfr_load_chsets(chgset_ctx_list_t *chgsets, zone_t *zone,
-                            const knot_rrset_t *their_soa)
+static int ixfr_load_chsets(journal_read_t **journal_read, zone_t *zone,
+                            const zone_contents_t *contents, const knot_rrset_t *their_soa)
 {
-	assert(chgsets);
+	assert(journal_read);
 	assert(zone);
 
 	/* Compare serials. */
-	uint32_t serial_to = zone_contents_serial(zone->contents);
+	uint32_t serial_to = zone_contents_serial(contents), j_serial_to;
 	uint32_t serial_from = knot_soa_serial(their_soa->rrs.rdata);
 	if (serial_compare(serial_to, serial_from) & SERIAL_MASK_LEQ) { /* We have older/same age zone. */
 		return KNOT_EUPTODATE;
 	}
 
-	return zone_chgset_ctx_load(conf(), zone, chgsets, serial_from);
+	zone_journal_t j = zone_journal(zone);
+	bool j_exists = false;
+	int ret = journal_info(j, &j_exists, NULL, NULL, &j_serial_to, NULL, NULL, NULL, NULL);
+	if (ret != KNOT_EOK) {
+		return ret;
+	} else if (!j_exists) {
+		return KNOT_ENOENT;
+	}
+
+	// please note that the journal serial_to might differ from zone SOA serial
+	// it is beacuse RCU lock is made at different moment than LMDB txn begin
+	return journal_read_begin(zone_journal(zone), false, serial_from, journal_read);
 }
 
 static int ixfr_query_check(knotd_qdata_t *qdata)
 {
 	NS_NEED_ZONE(qdata, KNOT_RCODE_NOTAUTH);
-	NS_NEED_AUTH(qdata, qdata->extra->zone->name, ACL_ACTION_TRANSFER);
+	NS_NEED_AUTH(qdata, ACL_ACTION_TRANSFER);
 	NS_NEED_ZONE_CONTENTS(qdata, KNOT_RCODE_SERVFAIL);
 
 	/* Need SOA authority record. */
@@ -134,19 +145,19 @@ static void ixfr_answer_cleanup(knotd_qdata_t *qdata)
 	struct ixfr_proc *ixfr = (struct ixfr_proc *)qdata->extra->ext;
 	knot_mm_t *mm = qdata->mm;
 
+	knot_rrset_clear(&ixfr->cur_rr, NULL);
 	ptrlist_free(&ixfr->proc.nodes, mm);
-	chgset_ctx_list_close(&ixfr->changesets);
+	journal_read_end(ixfr->journal_ctx);
 	mm_free(mm, qdata->extra->ext);
 
 	/* Allow zone changes (finished). */
 	rcu_read_unlock();
 }
 
-static int ixfr_answer_init(knotd_qdata_t *qdata)
+static int ixfr_answer_init(knotd_qdata_t *qdata, uint32_t *serial_from)
 {
 	assert(qdata);
 
-	/* Check IXFR query validity. */
 	if (ixfr_query_check(qdata) == KNOT_STATE_FAIL) {
 		if (qdata->rcode == KNOT_RCODE_FORMERR) {
 			return KNOT_EMALF;
@@ -155,11 +166,10 @@ static int ixfr_answer_init(knotd_qdata_t *qdata)
 		}
 	}
 
-	/* Compare serials. */
 	const knot_pktsection_t *authority = knot_pkt_section(qdata->query, KNOT_AUTHORITY);
 	const knot_rrset_t *their_soa = knot_pkt_rr(authority, 0);
+	*serial_from = knot_soa_serial(their_soa->rrs.rdata);
 
-	/* Initialize transfer processing. */
 	knot_mm_t *mm = qdata->mm;
 	struct ixfr_proc *xfer = mm_alloc(mm, sizeof(struct ixfr_proc));
 	if (xfer == NULL) {
@@ -167,7 +177,8 @@ static int ixfr_answer_init(knotd_qdata_t *qdata)
 	}
 	memset(xfer, 0, sizeof(struct ixfr_proc));
 
-	int ret = ixfr_load_chsets(&xfer->changesets, (zone_t *)qdata->extra->zone, their_soa);
+	int ret = ixfr_load_chsets(&xfer->journal_ctx, (zone_t *)qdata->extra->zone,
+	                           qdata->extra->contents, their_soa);
 	if (ret != KNOT_EOK) {
 		mm_free(mm, xfer);
 		return ret;
@@ -179,20 +190,11 @@ static int ixfr_answer_init(knotd_qdata_t *qdata)
 	knot_rrset_init_empty(&xfer->cur_rr);
 	xfer->qdata = qdata;
 
-	/* Put all changesets to processing queue. */
-	chgset_ctx_t *chs = NULL;
-	WALK_LIST(chs, xfer->changesets.l) {
-		ptrlist_add(&xfer->proc.nodes, chs, mm);
-		chgset_ctx_iterate(chs); // init already, so we don't have to decide later, no harm
-	}
+	ptrlist_add(&xfer->proc.nodes, xfer->journal_ctx, mm);
 
-	/* Keep first and last serial. */
-	chs = HEAD(xfer->changesets.l);
-	xfer->soa_from = chs->serial_from;
-	chs = TAIL(xfer->changesets.l);
-	xfer->soa_to = chs->serial_to;
+	xfer->soa_from = knot_soa_serial(their_soa->rrs.rdata);
+	xfer->soa_to = zone_contents_serial(qdata->extra->contents);
 
-	/* Set up cleanup callback. */
 	qdata->extra->ext = xfer;
 	qdata->extra->ext_cleanup = &ixfr_answer_cleanup;
 
@@ -220,7 +222,7 @@ static int ixfr_answer_soa(knot_pkt_t *pkt, knotd_qdata_t *qdata)
 	}
 
 	/* Guaranteed to have zone contents. */
-	const zone_node_t *apex = qdata->extra->zone->contents->apex;
+	const zone_node_t *apex = qdata->extra->contents->apex;
 	knot_rrset_t soa_rr = node_rrset(apex, KNOT_RRTYPE_SOA);
 	if (knot_rrset_empty(&soa_rr)) {
 		return KNOT_STATE_FAIL;
@@ -248,7 +250,8 @@ int ixfr_process_query(knot_pkt_t *pkt, knotd_qdata_t *qdata)
 	/* Initialize on first call. */
 	struct ixfr_proc *ixfr = qdata->extra->ext;
 	if (ixfr == NULL) {
-		int ret = ixfr_answer_init(qdata);
+		uint32_t soa_from = 0;
+		int ret = ixfr_answer_init(qdata, &soa_from);
 		ixfr = qdata->extra->ext;
 		switch (ret) {
 		case KNOT_EOK:       /* OK */
@@ -256,11 +259,11 @@ int ixfr_process_query(knot_pkt_t *pkt, knotd_qdata_t *qdata)
 				    ixfr->soa_from, ixfr->soa_to);
 			break;
 		case KNOT_EUPTODATE: /* Our zone is same age/older, send SOA. */
-			IXFROUT_LOG(LOG_INFO, qdata, "zone is up-to-date");
+			IXFROUT_LOG(LOG_INFO, qdata, "zone is up-to-date, serial %u", soa_from);
 			return ixfr_answer_soa(pkt, qdata);
 		case KNOT_ERANGE:    /* No history -> AXFR. */
 		case KNOT_ENOENT:
-			IXFROUT_LOG(LOG_INFO, qdata, "incomplete history, fallback to AXFR");
+			IXFROUT_LOG(LOG_INFO, qdata, "incomplete history, serial %u, fallback to AXFR", soa_from);
 			qdata->type = KNOTD_QUERY_TYPE_AXFR; /* Solve as AXFR. */
 			return axfr_process_query(pkt, qdata);
 		case KNOT_EDENIED:  /* Not authorized, already logged. */
@@ -282,7 +285,7 @@ int ixfr_process_query(knot_pkt_t *pkt, knotd_qdata_t *qdata)
 	}
 
 	/* Answer current packet (or continue). */
-	ret = xfr_process_list(pkt, &ixfr_process_changeset, qdata);
+	ret = xfr_process_list(pkt, &ixfr_process_journal, qdata);
 	switch (ret) {
 	case KNOT_ESPACE: /* Couldn't write more, send packet and continue. */
 		return KNOT_STATE_PRODUCE; /* Check for more. */
