@@ -1,4 +1,4 @@
-/*  Copyright (C) 2019 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2020 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,7 +11,7 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #include <assert.h>
@@ -68,7 +68,7 @@ void query_plan_free(struct query_plan *plan)
 	}
 
 	for (unsigned i = 0; i < KNOTD_STAGES; ++i) {
-		struct query_step *step = NULL, *next = NULL;
+		struct query_step *step, *next;
 		WALK_LIST_DELSAFE(step, next, plan->stage[i]) {
 			free(step);
 		}
@@ -123,10 +123,10 @@ int knotd_mod_in_hook(knotd_mod_t *mod, knotd_stage_t stage, knotd_mod_in_hook_f
 	return query_plan_step(mod->plan, stage, hook, mod);
 }
 
-knotd_mod_t *query_module_open(conf_t *conf, conf_mod_id_t *mod_id,
+knotd_mod_t *query_module_open(conf_t *conf, server_t *server, conf_mod_id_t *mod_id,
                                struct query_plan *plan, const knot_dname_t *zone)
 {
-	if (conf == NULL || mod_id == NULL || plan == NULL) {
+	if (conf == NULL || server == NULL || mod_id == NULL || plan == NULL) {
 		return NULL;
 	}
 
@@ -145,6 +145,7 @@ knotd_mod_t *query_module_open(conf_t *conf, conf_mod_id_t *mod_id,
 
 	module->plan = plan;
 	module->config = conf;
+	module->server = server;
 	module->zone = zone;
 	module->id = mod_id;
 	module->api = mod->api;
@@ -164,8 +165,10 @@ void query_module_close(knotd_mod_t *module)
 	zone_sign_ctx_free(module->sign_ctx);
 	free_zone_keys(module->keyset);
 	free(module->keyset);
-	kdnssec_ctx_deinit(module->dnssec);
-	free(module->dnssec);
+	if (module->dnssec != NULL) {
+		kdnssec_ctx_deinit(module->dnssec);
+		free(module->dnssec);
+	}
 
 	free(module);
 }
@@ -253,8 +256,6 @@ int knotd_mod_stats_add(knotd_mod_t *mod, const char *ctr_name, uint32_t idx_cou
 		stats += mod->stats_count;
 	}
 
-	mod->stats_count++;
-
 	if (idx_count == 1) {
 		stats->counter = 0;
 	} else {
@@ -269,6 +270,8 @@ int knotd_mod_stats_add(knotd_mod_t *mod, const char *ctr_name, uint32_t idx_cou
 	stats->name = ctr_name;
 	stats->count = idx_count;
 
+	mod->stats_count++;
+
 	return KNOT_EOK;
 }
 
@@ -340,10 +343,10 @@ knotd_conf_t knotd_conf_env(knotd_mod_t *mod, knotd_conf_env_t env)
 		out.single.string = config->hostname;
 		break;
 	case KNOTD_CONF_ENV_WORKERS_UDP:
-		out.single.integer = conf_udp_threads(config);
+		out.single.integer = config->cache.srv_udp_threads;
 		break;
 	case KNOTD_CONF_ENV_WORKERS_TCP:
-		out.single.integer = conf_tcp_threads(config);
+		out.single.integer = config->cache.srv_tcp_threads;
 		break;
 	default:
 		return out;
@@ -511,15 +514,11 @@ bool knotd_conf_addr_range_match(const knotd_conf_t *range,
 	for (size_t i = 0; i < range->count; i++) {
 		knotd_conf_val_t *val = &range->multi[i];
 		if (val->addr_max.ss_family == AF_UNSPEC) {
-			if (sockaddr_net_match((struct sockaddr *)addr,
-			                       (struct sockaddr *)&val->addr,
-			                       val->addr_mask)) {
+			if (sockaddr_net_match(addr, &val->addr, val->addr_mask)) {
 				return true;
 			}
 		} else {
-			if (sockaddr_range_match((struct sockaddr *)addr,
-			                         (struct sockaddr *)&val->addr,
-			                         (struct sockaddr *)&val->addr_max)) {
+			if (sockaddr_range_match(addr, &val->addr, &val->addr_max)) {
 				return true;
 			}
 		}
@@ -555,28 +554,30 @@ const knot_dname_t *knotd_qdata_zone_name(knotd_qdata_t *qdata)
 _public_
 knot_rrset_t knotd_qdata_zone_apex_rrset(knotd_qdata_t *qdata, uint16_t type)
 {
-	if (qdata == NULL || qdata->extra->zone == NULL ||
-	    qdata->extra->zone->contents == NULL) {
+	if (qdata == NULL || qdata->extra->contents == NULL) {
 		return node_rrset(NULL, type);
 	}
 
-	return node_rrset(qdata->extra->zone->contents->apex, type);
+	return node_rrset(qdata->extra->contents->apex, type);
 }
 
 _public_
 int knotd_mod_dnssec_init(knotd_mod_t *mod)
 {
-	if (mod == NULL) {
+	if (mod == NULL || mod->dnssec != NULL) {
 		return KNOT_EINVAL;
 	}
 
+	knot_lmdb_db_t *kaspdb = &mod->server->kaspdb;
+	kasp_db_ensure_init(kaspdb, mod->config); // probably redundant
+
 	mod->dnssec = calloc(1, sizeof(*(mod->dnssec)));
 	if (mod->dnssec == NULL) {
 		return KNOT_ENOMEM;
 	}
 
 	conf_val_t conf = conf_zone_get(mod->config, C_DNSSEC_SIGNING, mod->zone);
-	int ret = kdnssec_ctx_init(mod->config, mod->dnssec, mod->zone,
+	int ret = kdnssec_ctx_init(mod->config, mod->dnssec, mod->zone, kaspdb,
 	                           conf_bool(&conf) ? NULL : mod->id);
 	if (ret != KNOT_EOK) {
 		free(mod->dnssec);
