@@ -1,4 +1,4 @@
-/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2019 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,22 +11,25 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #include "knot/common/log.h"
 #include "knot/dnssec/zone-events.h"
 #include "knot/updates/zone-update.h"
+#include "knot/zone/adds_tree.h"
+#include "knot/zone/adjust.h"
 #include "knot/zone/serial.h"
 #include "knot/zone/zone-diff.h"
-#include "contrib/mempattern.h"
 #include "contrib/trim.h"
 #include "contrib/ucw/lists.h"
-#include "contrib/ucw/mempool.h"
 
 #include <urcu.h>
 
-static int init_incremental(zone_update_t *update, zone_t *zone, zone_contents_t *old_contents, bool deep_copy)
+// Call mem_trim() whenever accumuled size of updated zones reaches this size.
+#define UPDATE_MEMTRIM_AT (10 * 1024 * 1024)
+
+static int init_incremental(zone_update_t *update, zone_t *zone, zone_contents_t *old_contents)
 {
 	if (old_contents == NULL) {
 		return KNOT_EINVAL;
@@ -37,11 +40,9 @@ static int init_incremental(zone_update_t *update, zone_t *zone, zone_contents_t
 		return ret;
 	}
 
-	if (deep_copy) {
-		update->new_cont_deep_copy = true;
+	if (update->flags & UPDATE_HYBRID) {
 		update->new_cont = old_contents;
 	} else {
-		update->new_cont_deep_copy = false;
 		ret = apply_prepare_zone_copy(old_contents, &update->new_cont);
 		if (ret != KNOT_EOK) {
 			changeset_clear(&update->change);
@@ -49,8 +50,13 @@ static int init_incremental(zone_update_t *update, zone_t *zone, zone_contents_t
 		}
 	}
 
-	uint32_t apply_flags = update->flags & UPDATE_STRICT ? APPLY_STRICT : 0;
-	apply_init_ctx(update->a_ctx, update->new_cont, apply_flags);
+	uint32_t apply_flags = (update->flags & UPDATE_STRICT) ? APPLY_STRICT : 0;
+	apply_flags |= (update->flags & UPDATE_HYBRID) ? APPLY_UNIFY_FULL : 0;
+	ret = apply_init_ctx(update->a_ctx, update->new_cont, apply_flags);
+	if (ret != KNOT_EOK) {
+		changeset_clear(&update->change);
+		return ret;
+	}
 
 	/* Copy base SOA RR. */
 	update->change.soa_from =
@@ -66,14 +72,16 @@ static int init_incremental(zone_update_t *update, zone_t *zone, zone_contents_t
 
 static int init_full(zone_update_t *update, zone_t *zone)
 {
-	update->new_cont = zone_contents_new(zone->name);
+	update->new_cont = zone_contents_new(zone->name, true);
 	if (update->new_cont == NULL) {
 		return KNOT_ENOMEM;
 	}
 
-	update->new_cont_deep_copy = true;
-
-	apply_init_ctx(update->a_ctx, update->new_cont, 0);
+	int ret = apply_init_ctx(update->a_ctx, update->new_cont, APPLY_UNIFY_FULL);
+	if (ret != KNOT_EOK) {
+		zone_contents_free(update->new_cont);
+		return ret;
+	}
 
 	return KNOT_EOK;
 }
@@ -100,17 +108,15 @@ static int replace_soa(zone_contents_t *contents, const knot_rrset_t *rr)
 	return ret;
 }
 
-int init_base(zone_update_t *update, zone_t *zone, zone_contents_t *old_contents,
-              zone_update_flags_t flags)
+static int init_base(zone_update_t *update, zone_t *zone, zone_contents_t *old_contents,
+                     zone_update_flags_t flags)
 {
-	if (update == NULL || zone == NULL || (old_contents == NULL && (flags & UPDATE_INCREMENTAL))) {
+	if (update == NULL || zone == NULL) {
 		return KNOT_EINVAL;
 	}
 
 	memset(update, 0, sizeof(*update));
 	update->zone = zone;
-
-	mm_ctx_mempool(&update->mm, MM_DEFAULT_BLKSIZE);
 	update->flags = flags;
 
 	update->a_ctx = calloc(1, sizeof(*update->a_ctx));
@@ -118,13 +124,25 @@ int init_base(zone_update_t *update, zone_t *zone, zone_contents_t *old_contents
 		return KNOT_ENOMEM;
 	}
 
+	if (zone->control_update != NULL && zone->control_update != update) {
+		log_zone_warning(zone->name, "blocked zone update due to open control transaction");
+	}
+
+	knot_sem_wait(&zone->cow_lock);
+	update->a_ctx->cow_mutex = &zone->cow_lock;
+
+	if (old_contents == NULL) {
+		old_contents = zone->contents; // don't obtain this pointer before any other zone_update ceased to exist!
+	}
+
 	int ret = KNOT_EINVAL;
-	if (flags & UPDATE_INCREMENTAL) {
-		ret = init_incremental(update, zone, old_contents, flags & UPDATE_JOURNAL);
+	if (flags & (UPDATE_INCREMENTAL | UPDATE_HYBRID)) {
+		ret = init_incremental(update, zone, old_contents);
 	} else if (flags & UPDATE_FULL) {
 		ret = init_full(update, zone);
 	}
 	if (ret != KNOT_EOK) {
+		knot_sem_post(&zone->cow_lock);
 		free(update->a_ctx);
 	}
 
@@ -135,14 +153,14 @@ int init_base(zone_update_t *update, zone_t *zone, zone_contents_t *old_contents
 
 int zone_update_init(zone_update_t *update, zone_t *zone, zone_update_flags_t flags)
 {
-	return init_base(update, zone, zone->contents, flags);
+	return init_base(update, zone, NULL, flags);
 }
 
 int zone_update_from_differences(zone_update_t *update, zone_t *zone, zone_contents_t *old_cont,
 				 zone_contents_t *new_cont, zone_update_flags_t flags, bool ignore_dnssec)
 {
 	if (update == NULL || zone == NULL || new_cont == NULL ||
-	    !(flags & UPDATE_INCREMENTAL) || (flags & UPDATE_FULL)) {
+	    !(flags & (UPDATE_INCREMENTAL | UPDATE_HYBRID)) || (flags & UPDATE_FULL)) {
 		return KNOT_EINVAL;
 	}
 
@@ -152,9 +170,20 @@ int zone_update_from_differences(zone_update_t *update, zone_t *zone, zone_conte
 		return ret;
 	}
 
+	ret = init_base(update, zone, old_cont, flags);
+	if (ret != KNOT_EOK) {
+		changeset_clear(&diff);
+		return ret;
+	}
+
+	if (old_cont == NULL) {
+		old_cont = zone->contents;
+	}
+
 	ret = zone_contents_diff(old_cont, new_cont, &diff, ignore_dnssec);
 	if (ret != KNOT_EOK && ret != KNOT_ENODIFF && ret != KNOT_ESEMCHECK) {
 		changeset_clear(&diff);
+		zone_update_clear(update);
 		return ret;
 	}
 
@@ -162,12 +191,6 @@ int zone_update_from_differences(zone_update_t *update, zone_t *zone, zone_conte
 	// remained the same and has to be incremented.
 	bool diff_semcheck = (ret == KNOT_ESEMCHECK);
 
-	ret = init_base(update, zone, old_cont, flags);
-	if (ret != KNOT_EOK) {
-		changeset_clear(&diff);
-		return ret;
-	}
-
 	ret = zone_update_apply_changeset(update, &diff);
 	changeset_clear(&diff);
 	if (ret != KNOT_EOK) {
@@ -184,6 +207,7 @@ int zone_update_from_differences(zone_update_t *update, zone_t *zone, zone_conte
 		log_zone_info(zone->name, "automatic SOA serial increment");
 	}
 
+	update->init_cont = new_cont;
 	return KNOT_EOK;
 }
 
@@ -196,22 +220,27 @@ int zone_update_from_contents(zone_update_t *update, zone_t *zone_without_conten
 
 	memset(update, 0, sizeof(*update));
 	update->zone = zone_without_contents;
-
-	mm_ctx_mempool(&update->mm, MM_DEFAULT_BLKSIZE);
 	update->flags = flags;
-
 	update->new_cont = new_cont;
-	update->new_cont_deep_copy = true;
 
 	update->a_ctx = calloc(1, sizeof(*update->a_ctx));
 	if (update->a_ctx == NULL) {
 		return KNOT_ENOMEM;
 	}
 
-	if (flags & UPDATE_INCREMENTAL) {
+	if (zone_without_contents->control_update != NULL) {
+		log_zone_warning(zone_without_contents->name,
+		                 "blocked zone update due to open control transaction");
+	}
+
+	knot_sem_wait(&update->zone->cow_lock);
+	update->a_ctx->cow_mutex = &update->zone->cow_lock;
+
+	if (flags & (UPDATE_INCREMENTAL | UPDATE_HYBRID)) {
 		int ret = changeset_init(&update->change, zone_without_contents->name);
 		if (ret != KNOT_EOK) {
 			free(update->a_ctx);
+			knot_sem_post(&update->zone->cow_lock);
 			return ret;
 		}
 
@@ -219,37 +248,75 @@ int zone_update_from_contents(zone_update_t *update, zone_t *zone_without_conten
 		if (update->change.soa_from == NULL) {
 			changeset_clear(&update->change);
 			free(update->a_ctx);
+			knot_sem_post(&update->zone->cow_lock);
 			return KNOT_ENOMEM;
 		}
 	}
 
-	uint32_t apply_flags = update->flags & UPDATE_STRICT ? APPLY_STRICT : 0;
-	apply_init_ctx(update->a_ctx, update->new_cont, apply_flags);
+	uint32_t apply_flags = (update->flags & UPDATE_STRICT) ? APPLY_STRICT : 0;
+	int ret = apply_init_ctx(update->a_ctx, update->new_cont, apply_flags | APPLY_UNIFY_FULL);
+	if (ret != KNOT_EOK) {
+		changeset_clear(&update->change);
+		free(update->a_ctx);
+		knot_sem_post(&update->zone->cow_lock);
+		return ret;
+	}
 
 	return KNOT_EOK;
 }
 
+int zone_update_start_extra(zone_update_t *update)
+{
+	assert((update->flags & (UPDATE_INCREMENTAL | UPDATE_HYBRID)));
+
+	int ret = changeset_init(&update->extra_ch, update->new_cont->apex->owner);
+	if (ret != KNOT_EOK) {
+		return ret;
+	}
+
+	if (update->init_cont != NULL) {
+		ret = zone_update_increment_soa(update, conf());
+		if (ret != KNOT_EOK) {
+			return ret;
+		}
+
+		ret = zone_contents_diff(update->init_cont, update->new_cont, &update->extra_ch, false);
+		if (ret != KNOT_EOK) {
+			return ret;
+		}
+	} else {
+		update->extra_ch.soa_from = node_create_rrset(update->new_cont->apex, KNOT_RRTYPE_SOA);
+		if (update->extra_ch.soa_from == NULL) {
+			return KNOT_ENOMEM;
+		}
+
+		ret = zone_update_increment_soa(update, conf());
+		if (ret != KNOT_EOK) {
+			return ret;
+		}
+
+		update->extra_ch.soa_to = node_create_rrset(update->new_cont->apex, KNOT_RRTYPE_SOA);
+		if (update->extra_ch.soa_to == NULL) {
+			return KNOT_ENOMEM;
+		}
+	}
+
+	update->flags |= UPDATE_EXTRA_CHSET;
+	return KNOT_EOK;
+}
+
 const zone_node_t *zone_update_get_node(zone_update_t *update, const knot_dname_t *dname)
 {
 	if (update == NULL || dname == NULL) {
 		return NULL;
 	}
 
-	return zone_contents_find_node(update->new_cont, dname);
-}
-
-const zone_node_t *zone_update_get_apex(zone_update_t *update)
-{
-	if (update == NULL) {
-		return NULL;
-	}
-
-	return zone_update_get_node(update, update->zone->name);
+	return zone_contents_node_or_nsec3(update->new_cont, dname);
 }
 
 uint32_t zone_update_current_serial(zone_update_t *update)
 {
-	const zone_node_t *apex = zone_update_get_apex(update);
+	const zone_node_t *apex = update->new_cont->apex;
 	if (apex != NULL) {
 		return knot_soa_serial(node_rdataset(apex, KNOT_RRTYPE_SOA)->rdata);
 	} else {
@@ -257,13 +324,24 @@ uint32_t zone_update_current_serial(zone_update_t *update)
 	}
 }
 
+bool zone_update_changed_nsec3param(const zone_update_t *update)
+{
+	if (update->zone->contents == NULL) {
+		return true;
+	}
+
+	dnssec_nsec3_params_t *orig = &update->zone->contents->nsec3_params;
+	dnssec_nsec3_params_t *upd = &update->new_cont->nsec3_params;
+	return !dnssec_nsec3_params_match(orig, upd);
+}
+
 const knot_rdataset_t *zone_update_from(zone_update_t *update)
 {
 	if (update == NULL) {
 		return NULL;
 	}
 
-	if (update->flags & UPDATE_INCREMENTAL) {
+	if (update->flags & (UPDATE_INCREMENTAL | UPDATE_HYBRID)) {
 		const zone_node_t *apex = update->zone->contents->apex;
 		return node_rdataset(apex, KNOT_RRTYPE_SOA);
 	}
@@ -280,7 +358,7 @@ const knot_rdataset_t *zone_update_to(zone_update_t *update)
 	if (update->flags & UPDATE_FULL) {
 		const zone_node_t *apex = update->new_cont->apex;
 		return node_rdataset(apex, KNOT_RRTYPE_SOA);
-	} else if (update->flags & UPDATE_INCREMENTAL) {
+	} else {
 		if (update->change.soa_to == NULL) {
 			return NULL;
 		}
@@ -292,58 +370,114 @@ const knot_rdataset_t *zone_update_to(zone_update_t *update)
 
 void zone_update_clear(zone_update_t *update)
 {
-	if (update == NULL) {
+	if (update == NULL || update->zone == NULL) {
 		return;
 	}
 
-	if (update->flags & UPDATE_INCREMENTAL) {
-		/* Revert any changes on error, do nothing on success. */
-		if (update->new_cont_deep_copy) {
-			update_cleanup(update->a_ctx);
-			zone_contents_deep_free(update->new_cont);
-		} else {
-			update_rollback(update->a_ctx);
-			update_free_zone(update->new_cont);
-		}
+	if (update->flags & (UPDATE_INCREMENTAL | UPDATE_HYBRID)) {
 		changeset_clear(&update->change);
-	} else if (update->flags & UPDATE_FULL) {
-		assert(update->new_cont_deep_copy);
-		zone_contents_deep_free(update->new_cont);
+		changeset_clear(&update->extra_ch);
 	}
+
+	zone_contents_deep_free(update->init_cont);
+
+	if (update->flags & (UPDATE_FULL | UPDATE_HYBRID)) {
+		apply_cleanup(update->a_ctx);
+		zone_contents_deep_free(update->new_cont);
+	} else {
+		apply_rollback(update->a_ctx);
+	}
+
 	free(update->a_ctx);
-	mp_delete(update->mm.ctx);
 	memset(update, 0, sizeof(*update));
 }
 
+inline static void update_affected_rrtype(zone_update_t *update, uint16_t rrtype)
+{
+	switch (rrtype) {
+	case KNOT_RRTYPE_NSEC:
+	case KNOT_RRTYPE_NSEC3:
+		update->flags |= UPDATE_CHANGED_NSEC;
+		break;
+	}
+}
+
+static int solve_add_different_ttl(zone_update_t *update, const knot_rrset_t *add)
+{
+	if (add->type == KNOT_RRTYPE_RRSIG || add->type == KNOT_RRTYPE_SOA) {
+		return KNOT_EOK;
+	}
+
+	const zone_node_t *exist_node = zone_contents_find_node(update->new_cont, add->owner);
+	const knot_rrset_t exist_rr = node_rrset(exist_node, add->type);
+	if (knot_rrset_empty(&exist_rr) || exist_rr.ttl == add->ttl) {
+		return KNOT_EOK;
+	}
+
+	knot_dname_txt_storage_t buff;
+	char *owner = knot_dname_to_str(buff, add->owner, sizeof(buff));
+	if (owner == NULL) {
+		owner = "";
+	}
+	char type[16] = "";
+	knot_rrtype_to_string(add->type, type, sizeof(type));
+	log_zone_notice(update->zone->name, "TTL mismatch, owner %s, type %s, "
+	                "TTL set to %u", owner, type, add->ttl);
+
+	knot_rrset_t *exist_copy = knot_rrset_copy(&exist_rr, NULL);
+	if (exist_copy == NULL) {
+		return KNOT_ENOMEM;
+	}
+	int ret = zone_update_remove(update, exist_copy);
+	if (ret == KNOT_EOK) {
+		exist_copy->ttl = add->ttl;
+		ret = zone_update_add(update, exist_copy);
+	}
+	knot_rrset_free(exist_copy, NULL);
+	return ret;
+}
+
 int zone_update_add(zone_update_t *update, const knot_rrset_t *rrset)
 {
 	if (update == NULL || rrset == NULL) {
 		return KNOT_EINVAL;
 	}
+	if (knot_rrset_empty(rrset)) {
+		return KNOT_EOK;
+	}
 
-	if (update->flags & UPDATE_INCREMENTAL) {
-		int ret = changeset_add_addition(&update->change, rrset, CHANGESET_CHECK);
+	if (update->flags & (UPDATE_INCREMENTAL | UPDATE_HYBRID)) {
+		int ret = solve_add_different_ttl(update, rrset);
+		if (ret == KNOT_EOK) {
+			ret = changeset_add_addition(&update->change, rrset, CHANGESET_CHECK);
+		}
+		if (ret == KNOT_EOK && (update->flags & UPDATE_EXTRA_CHSET)) {
+			ret = changeset_add_addition(&update->extra_ch, rrset, CHANGESET_CHECK);
+		}
 		if (ret != KNOT_EOK) {
 			return ret;
 		}
+	}
 
+	if (update->flags & UPDATE_INCREMENTAL) {
 		if (rrset->type == KNOT_RRTYPE_SOA) {
-			/* replace previous SOA */
-			ret = apply_replace_soa(update->a_ctx, &update->change);
+			// replace previous SOA
+			int ret = apply_replace_soa(update->a_ctx, rrset);
 			if (ret != KNOT_EOK) {
 				changeset_remove_addition(&update->change, rrset);
 			}
 			return ret;
 		}
 
-		ret = apply_add_rr(update->a_ctx, rrset);
+		int ret = apply_add_rr(update->a_ctx, rrset);
 		if (ret != KNOT_EOK) {
 			changeset_remove_addition(&update->change, rrset);
 			return ret;
 		}
 
+		update_affected_rrtype(update, rrset->type);
 		return KNOT_EOK;
-	} else if (update->flags & UPDATE_FULL) {
+	} else if (update->flags & (UPDATE_FULL | UPDATE_HYBRID)) {
 		if (rrset->type == KNOT_RRTYPE_SOA) {
 			/* replace previous SOA */
 			return replace_soa(update->new_cont, rrset);
@@ -352,12 +486,12 @@ int zone_update_add(zone_update_t *update, const knot_rrset_t *rrset)
 		zone_node_t *n = NULL;
 		int ret = zone_contents_add_rr(update->new_cont, rrset, &n);
 		if (ret == KNOT_ETTL) {
-			char buff[KNOT_DNAME_TXT_MAXLEN + 1];
+			knot_dname_txt_storage_t buff;
 			char *owner = knot_dname_to_str(buff, rrset->owner, sizeof(buff));
 			if (owner == NULL) {
 				owner = "";
 			}
-			char type[16] = { '\0' };
+			char type[16] = "";
 			knot_rrtype_to_string(rrset->type, type, sizeof(type));
 			log_zone_notice(update->new_cont->apex->owner,
 			                "TTL mismatch, owner %s, type %s, "
@@ -377,30 +511,33 @@ int zone_update_remove(zone_update_t *update, const knot_rrset_t *rrset)
 		return KNOT_EINVAL;
 	}
 
-	if (update->flags & UPDATE_INCREMENTAL) {
+	if ((update->flags & (UPDATE_INCREMENTAL | UPDATE_HYBRID)) && rrset->type != KNOT_RRTYPE_SOA) {
 		int ret = changeset_add_removal(&update->change, rrset, CHANGESET_CHECK);
+		if (ret == KNOT_EOK && (update->flags & UPDATE_EXTRA_CHSET)) {
+			ret = changeset_add_removal(&update->extra_ch, rrset, CHANGESET_CHECK);
+		}
 		if (ret != KNOT_EOK) {
 			return ret;
 		}
+	}
 
+	if (update->flags & UPDATE_INCREMENTAL) {
 		if (rrset->type == KNOT_RRTYPE_SOA) {
 			/* SOA is replaced with addition */
 			return KNOT_EOK;
 		}
 
-		ret = apply_remove_rr(update->a_ctx, rrset);
+		int ret = apply_remove_rr(update->a_ctx, rrset);
 		if (ret != KNOT_EOK) {
 			changeset_remove_removal(&update->change, rrset);
 			return ret;
 		}
 
+		update_affected_rrtype(update, rrset->type);
 		return KNOT_EOK;
-	} else if (update->flags & UPDATE_FULL) {
+	} else if (update->flags & (UPDATE_FULL | UPDATE_HYBRID)) {
 		zone_node_t *n = NULL;
-		knot_rrset_t *rrs_copy = knot_rrset_copy(rrset, &update->mm);
-		int ret = zone_contents_remove_rr(update->new_cont, rrs_copy, &n);
-		knot_rrset_free(rrs_copy, &update->mm);
-		return ret;
+		return zone_contents_remove_rr(update->new_cont, rrset, &n);
 	} else {
 		return KNOT_EINVAL;
 	}
@@ -412,40 +549,33 @@ int zone_update_remove_rrset(zone_update_t *update, knot_dname_t *owner, uint16_
 		return KNOT_EINVAL;
 	}
 
-	if (update->flags & UPDATE_INCREMENTAL) {
-		/* Remove the RRSet from the original node */
-		const zone_node_t *node = zone_contents_find_node(update->new_cont, owner);
-		if (node != NULL) {
-			knot_rrset_t rrset = node_rrset(node, type);
-			if (rrset.owner == NULL) {
-				return KNOT_ENOENT;
-			}
-			int ret = changeset_add_removal(&update->change, &rrset,
-			                                CHANGESET_CHECK);
-			if (ret != KNOT_EOK) {
-				return ret;
-			}
+	const zone_node_t *node = zone_contents_node_or_nsec3(update->new_cont, owner);
+	if (node == NULL) {
+		return KNOT_ENONODE;
+	}
 
-			if (type == KNOT_RRTYPE_SOA) {
-				/* SOA is replaced with addition */
-				return KNOT_EOK;
-			}
+	knot_rrset_t rrset = node_rrset(node, type);
+	if (rrset.owner == NULL) {
+		return KNOT_ENOENT;
+	}
 
-			ret = apply_remove_rr(update->a_ctx, &rrset);
-			if (ret != KNOT_EOK) {
-				return ret;
-			}
-		} else {
-			return KNOT_ENONODE;
-		}
-	} else if (update->flags & UPDATE_FULL) {
-		/* Remove the RRSet from the non-synthesized new node */
-		const zone_node_t *node = zone_contents_find_node(update->new_cont, owner);
-		if (node == NULL) {
-			return KNOT_ENONODE;
-		}
+	return zone_update_remove(update, &rrset);
+}
 
-		knot_rrset_t rrset = node_rrset(node, type);
+int zone_update_remove_node(zone_update_t *update, const knot_dname_t *owner)
+{
+	if (update == NULL || owner == NULL) {
+		return KNOT_EINVAL;
+	}
+
+	const zone_node_t *node = zone_contents_node_or_nsec3(update->new_cont, owner);
+	if (node == NULL) {
+		return KNOT_ENONODE;
+	}
+
+	size_t rrset_count = node->rrset_count;
+	for (int i = 0; i < rrset_count; ++i) {
+		knot_rrset_t rrset = node_rrset_at(node, rrset_count - 1 - i);
 		int ret = zone_update_remove(update, &rrset);
 		if (ret != KNOT_EOK) {
 			return ret;
@@ -455,80 +585,19 @@ int zone_update_remove_rrset(zone_update_t *update, knot_dname_t *owner, uint16_
 	return KNOT_EOK;
 }
 
-int zone_update_remove_node(zone_update_t *update, const knot_dname_t *owner)
+static int update_chset_step(const knot_rrset_t *rrset, bool addition, void *ctx)
 {
-	if (update == NULL || owner == NULL) {
-		return KNOT_EINVAL;
+	zone_update_t *update = ctx;
+	if (addition) {
+		return zone_update_add(update, rrset);
+	} else {
+		return zone_update_remove(update, rrset);
 	}
-
-	if (update->flags & UPDATE_INCREMENTAL) {
-		/* Remove all RRSets from the new node */
-		const zone_node_t *node = zone_contents_find_node(update->new_cont, owner);
-		if (node != NULL) {
-			size_t rrset_count = node->rrset_count;
-			for (int i = 0; i < rrset_count; ++i) {
-				knot_rrset_t rrset = node_rrset_at(node, rrset_count - 1 - i);
-				int ret = changeset_add_removal(&update->change, &rrset,
-				                                CHANGESET_CHECK);
-				if (ret != KNOT_EOK) {
-					return ret;
-				}
-
-				if (rrset.type == KNOT_RRTYPE_SOA) {
-					/* SOA is replaced with addition */
-					continue;
-				}
-
-				ret = apply_remove_rr(update->a_ctx, &rrset);
-				if (ret != KNOT_EOK) {
-					return ret;
-				}
-			}
-		} else {
-			return KNOT_ENONODE;
-		}
-	} else if (update->flags & UPDATE_FULL) {
-		/* Remove all RRSets from the non-synthesized new node */
-		const zone_node_t *node = zone_contents_find_node(update->new_cont, owner);
-		if (node == NULL) {
-			return KNOT_ENONODE;
-		}
-
-		size_t rrset_count = node->rrset_count;
-		for (int i = 0; i < rrset_count; ++i) {
-			knot_rrset_t rrset = node_rrset_at(node, rrset_count - 1 - i);
-			int ret = zone_update_remove(update, &rrset);
-			if (ret != KNOT_EOK) {
-				return ret;
-			}
-		}
-	}
-
-	return KNOT_EOK;
 }
 
 int zone_update_apply_changeset(zone_update_t *update, const changeset_t *changes)
 {
-	int ret = KNOT_EOK;
-	if (update->flags & UPDATE_INCREMENTAL) {
-		ret = changeset_merge(&update->change, changes, CHANGESET_CHECK_CANCELOUT);
-	}
-	if (ret == KNOT_EOK) {
-		ret = apply_changeset_directly(update->a_ctx, changes);
-	}
-	return ret;
-}
-
-int zone_update_apply_changeset_fix(zone_update_t *update, changeset_t *changes)
-{
-	int ret = changeset_cancelout(changes);
-	if (ret == KNOT_EOK) {
-		ret = changeset_preapply_fix(update->new_cont, changes);
-	}
-	if (ret == KNOT_EOK) {
-		ret = zone_update_apply_changeset(update, changes);
-	}
-	return ret;
+	return changeset_walk(changes, update_chset_step, update);
 }
 
 int zone_update_apply_changeset_reverse(zone_update_t *update, const changeset_t *changes)
@@ -545,7 +614,7 @@ static int set_new_soa(zone_update_t *update, unsigned serial_policy)
 {
 	assert(update);
 
-	knot_rrset_t *soa_cpy = node_create_rrset(zone_update_get_apex(update),
+	knot_rrset_t *soa_cpy = node_create_rrset(update->new_cont->apex,
 	                                          KNOT_RRTYPE_SOA);
 	if (soa_cpy == NULL) {
 		return KNOT_ENOMEM;
@@ -584,55 +653,45 @@ int zone_update_increment_soa(zone_update_t *update, conf_t *conf)
 	return set_new_soa(update, conf_opt(&val));
 }
 
-static int commit_incremental(conf_t *conf, zone_update_t *update,
-                              zone_contents_t **contents_out)
+static int commit_journal(conf_t *conf, zone_update_t *update)
+{
+	conf_val_t val = conf_zone_get(conf, C_JOURNAL_CONTENT, update->zone->name);
+	unsigned content = conf_opt(&val);
+	int ret = KNOT_EOK;
+	if ((update->flags & UPDATE_INCREMENTAL) ||
+	    (update->flags & UPDATE_HYBRID)) {
+		changeset_t *extra = (update->flags & UPDATE_EXTRA_CHSET) ? &update->extra_ch : NULL;
+		if (content != JOURNAL_CONTENT_NONE && !changeset_empty(&update->change)) {
+			ret = zone_change_store(conf, update->zone, &update->change, extra);
+		}
+	} else {
+		if (content == JOURNAL_CONTENT_ALL) {
+			return zone_in_journal_store(conf, update->zone, update->new_cont);
+		} else if (content != JOURNAL_CONTENT_NONE) { // zone_in_journal_store does this automatically
+			return zone_changes_clear(conf, update->zone);
+		}
+	}
+	return ret;
+}
+
+static int commit_incremental(conf_t *conf, zone_update_t *update)
 {
 	assert(update);
-	assert(contents_out);
 
-	if (changeset_empty(&update->change)) {
-		changeset_clear(&update->change);
-		if (update->zone->contents == NULL || update->new_cont_deep_copy) {
-			*contents_out = update->new_cont;
-		}
-		return KNOT_EOK;
-	}
-
-	zone_contents_t *new_contents = update->new_cont;
-	int ret = KNOT_EOK;
-	if (zone_update_to(update) == NULL) {
+	if (zone_update_to(update) == NULL && !changeset_empty(&update->change)) {
 		/* No SOA in the update, create one according to the current policy */
-		ret = zone_update_increment_soa(update, conf);
-		if (ret != KNOT_EOK) {
-			zone_update_clear(update);
-			return ret;
-		}
-	}
-
-	ret = apply_finalize(update->a_ctx);
-	if (ret != KNOT_EOK) {
-		zone_update_clear(update);
-		return ret;
-	}
-
-	/* Write changes to journal if all went well. */
-	conf_val_t val = conf_zone_get(conf, C_JOURNAL_CONTENT, update->zone->name);
-	if (conf_opt(&val) != JOURNAL_CONTENT_NONE) {
-		ret = zone_change_store(conf, update->zone, &update->change);
+		int ret = zone_update_increment_soa(update, conf);
 		if (ret != KNOT_EOK) {
 			return ret;
 		}
 	}
 
-	*contents_out = new_contents;
-
 	return KNOT_EOK;
 }
 
-static int commit_full(conf_t *conf, zone_update_t *update, zone_contents_t **contents_out)
+static int commit_full(conf_t *conf, zone_update_t *update)
 {
 	assert(update);
-	assert(contents_out);
 
 	/* Check if we have SOA. We might consider adding full semantic check here.
 	 * But if we wanted full sem-check I'd consider being it controlled by a flag
@@ -641,60 +700,66 @@ static int commit_full(conf_t *conf, zone_update_t *update, zone_contents_t **co
 		return KNOT_ESEMCHECK;
 	}
 
-	int ret = zone_contents_adjust_full(update->new_cont);
-	if (ret != KNOT_EOK) {
-		zone_update_clear(update);
-		return ret;
-	}
-
-	/* Store new zone contents in journal. */
-	conf_val_t val = conf_zone_get(conf, C_JOURNAL_CONTENT, update->zone->name);
-	unsigned content = conf_opt(&val);
-	if (content == JOURNAL_CONTENT_ALL) {
-		ret = zone_in_journal_store(conf, update->zone, update->new_cont);
-	} else if (content != JOURNAL_CONTENT_NONE) { // zone_in_journal_store does this automatically
-		ret = zone_changes_clear(conf, update->zone);
-	}
-
-	*contents_out = update->new_cont;
-
-	return ret;
+	return KNOT_EOK;
 }
 
-/*! \brief Routine for calling call_rcu() easier way.
+typedef struct {
+	pthread_mutex_t lock;
+	size_t counter;
+} counter_reach_t;
+
+static bool counter_reach(counter_reach_t *counter, size_t increment, size_t limit)
+{
+	bool reach = false;
+	pthread_mutex_lock(&counter->lock);
+	counter->counter += increment;
+	if (counter->counter >= limit) {
+		counter->counter = 0;
+		reach = true;
+	}
+	pthread_mutex_unlock(&counter->lock);
+	return reach;
+}
+
+/*! \brief Struct for what needs to be cleared after RCU.
  *
- * Consider moving elsewhere, as it has no direct relation to zone-update.
+ * This can't be zone_update_t structure as this might be already freed at that time.
  */
 typedef struct {
 	struct rcu_head rcuhead;
-	void (*callback)(void *);
-	void *ctx;
-	bool free_ctx;
-} callrcu_wrapper_t;
 
-static void callrcu_wrapper_cb(struct rcu_head *param)
+	zone_contents_t *free_contents;
+	void (*free_method)(zone_contents_t *);
+
+	apply_ctx_t *cleanup_apply;
+
+	size_t new_cont_size;
+} update_clear_ctx_t;
+
+static void update_clear(struct rcu_head *param)
 {
-	callrcu_wrapper_t *wrap = (callrcu_wrapper_t *)param;
-	wrap->callback(wrap->ctx);
-	if (wrap->free_ctx) {
-		free(wrap->ctx);
-	}
-	free(wrap);
+	static counter_reach_t counter = { PTHREAD_MUTEX_INITIALIZER, 0 };
 
-	// Trim extra heap.
-	mem_trim();
+	update_clear_ctx_t *ctx = (update_clear_ctx_t *)param;
+
+	ctx->free_method(ctx->free_contents);
+	apply_cleanup(ctx->cleanup_apply);
+	free(ctx->cleanup_apply);
+
+	if (counter_reach(&counter, ctx->new_cont_size, UPDATE_MEMTRIM_AT)) {
+		mem_trim();
+	}
+
+	free(ctx);
 }
 
-/* NOTE: Does nothing if not enough memory. */
-static void callrcu_wrapper(void *ctx, void *callback, bool free_ctx)
+static void discard_adds_tree(zone_update_t *update)
 {
-	callrcu_wrapper_t *wrap = calloc(1, sizeof(callrcu_wrapper_t));
-	if (wrap != NULL) {
-		wrap->callback = callback;
-		wrap->ctx = ctx;
-		wrap->free_ctx = free_ctx;
-		call_rcu((struct rcu_head *)wrap, callrcu_wrapper_cb);
+	additionals_tree_free(update->new_cont->adds_tree);
+	if (update->zone->contents != NULL) {
+		update->zone->contents->adds_tree = NULL;
 	}
+	update->new_cont->adds_tree = NULL;
 }
 
 int zone_update_commit(conf_t *conf, zone_update_t *update)
@@ -704,64 +769,98 @@ int zone_update_commit(conf_t *conf, zone_update_t *update)
 	}
 
 	int ret = KNOT_EOK;
-	zone_contents_t *new_contents = NULL;
+
 	if (update->flags & UPDATE_INCREMENTAL) {
-		ret = commit_incremental(conf, update, &new_contents);
+		if (changeset_empty(&update->change) &&
+		    update->zone->contents != NULL) {
+			changeset_clear(&update->change);
+			changeset_clear(&update->extra_ch);
+			zone_update_clear(update);
+			return KNOT_EOK;
+		}
+	}
+	if (update->flags & UPDATE_INCREMENTAL) {
+		ret = commit_incremental(conf, update);
+	} else if ((update->flags & UPDATE_HYBRID)) {
+		ret = commit_incremental(conf, update);
 	} else {
-		ret = commit_full(conf, update, &new_contents);
+		ret = commit_full(conf, update);
 	}
 	if (ret != KNOT_EOK) {
 		return ret;
 	}
 
-	/* If there is anything to change. */
-	if (new_contents == NULL) {
-		return KNOT_EOK;
+	if ((update->flags & (UPDATE_HYBRID | UPDATE_FULL))) {
+		ret = zone_adjust_full(update->new_cont);
+	} else {
+		ret = zone_adjust_incremental_update(update);
+	}
+	if (ret != KNOT_EOK) {
+		discard_adds_tree(update);
+		return ret;
 	}
 
 	/* Check the zone size. */
-	conf_val_t val = conf_zone_get(conf, C_MAX_ZONE_SIZE, update->zone->name);
+	conf_val_t val = conf_zone_get(conf, C_ZONE_MAX_SIZE, update->zone->name);
+	if (val.code != KNOT_EOK) {
+		val = conf_zone_get(conf, C_MAX_ZONE_SIZE, update->zone->name);
+	}
 	size_t size_limit = conf_int(&val);
 
-	if (new_contents->size > size_limit) {
-		/* Recoverable error. */
+	if (update->new_cont->size > size_limit) {
+		discard_adds_tree(update);
 		return KNOT_EZONESIZE;
 	}
 
+	ret = commit_journal(conf, update);
+	if (ret != KNOT_EOK) {
+		discard_adds_tree(update);
+		return ret;
+	}
+
 	/* Check if the zone was re-signed upon zone load to ensure proper flush
 	 * even if the SOA serial wasn't incremented by re-signing. */
 	val = conf_zone_get(conf, C_DNSSEC_SIGNING, update->zone->name);
 	bool dnssec = conf_bool(&val);
-	if (!changeset_empty(&update->change) && dnssec) {
-		update->zone->zonefile.resigned = true;
-	}
 
-	/* Abort control transaction if any. */
-	if (update->zone->control_update != NULL &&
-	    update->zone->control_update != update) {
-		log_zone_warning(update->zone->name, "control transaction aborted");
-		zone_control_clear(update->zone);
+	if (dnssec) {
+		update->zone->zonefile.resigned = true;
+
+		if (zone_is_slave(conf, update->zone)) {
+			ret = zone_set_lastsigned_serial(update->zone,
+			                                 zone_contents_serial(update->new_cont));
+			if (ret != KNOT_EOK) {
+				log_zone_warning(update->zone->name,
+				                 "unable to save lastsigned serial, "
+				                 "future transfers might be broken");
+			}
+		}
 	}
 
 	/* Switch zone contents. */
 	zone_contents_t *old_contents;
-	old_contents = zone_switch_contents(update->zone, new_contents);
+	old_contents = zone_switch_contents(update->zone, update->new_cont);
 
-	/* Sync RCU. */
-	if (update->flags & UPDATE_FULL) {
-		assert(update->new_cont_deep_copy);
-		callrcu_wrapper(old_contents, zone_contents_deep_free, false);
-	} else if (update->flags & UPDATE_INCREMENTAL) {
-		if (update->new_cont_deep_copy) {
-			callrcu_wrapper(old_contents, zone_contents_deep_free, false);
-		} else {
-			callrcu_wrapper(old_contents, update_free_zone, false);
-		}
+	if (update->flags & (UPDATE_INCREMENTAL | UPDATE_HYBRID)) {
 		changeset_clear(&update->change);
+		changeset_clear(&update->extra_ch);
+	}
+	zone_contents_deep_free(update->init_cont);
+
+	update_clear_ctx_t *clear_ctx = calloc(1, sizeof(*clear_ctx));
+	if (clear_ctx != NULL) {
+		clear_ctx->free_contents = old_contents;
+		clear_ctx->free_method = (
+			(update->flags & (UPDATE_FULL | UPDATE_HYBRID)) ?
+			zone_contents_deep_free : update_free_zone
+		);
+		clear_ctx->cleanup_apply = update->a_ctx;
+		clear_ctx->new_cont_size = update->new_cont->size;
+
+		call_rcu((struct rcu_head *)clear_ctx, update_clear);
+	} else {
+		log_zone_error(update->zone->name, "failed to deallocate unused memory");
 	}
-	callrcu_wrapper(update->a_ctx, update_cleanup, true);
-	update->a_ctx = NULL;
-	update->new_cont = NULL;
 
 	/* Sync zonefile immediately if configured. */
 	val = conf_zone_get(conf, C_ZONEFILE_SYNC, update->zone->name);
@@ -769,130 +868,18 @@ int zone_update_commit(conf_t *conf, zone_update_t *update)
 		zone_events_schedule_now(update->zone, ZONE_EVENT_FLUSH);
 	}
 
-	return KNOT_EOK;
-}
-
-static int iter_init_tree_iters(zone_update_iter_t *it, zone_update_t *update,
-                                bool nsec3)
-{
-	/* Set zone iterator. */
-	zone_contents_t *_contents = update->new_cont;
-
-	/* Begin iteration. We can safely assume _contents is a valid pointer. */
-	zone_tree_t *tree = nsec3 ? _contents->nsec3_nodes : _contents->nodes;
-	it->tree_it = trie_it_begin(tree);
-	if (it->tree_it == NULL) {
-		return KNOT_ENOMEM;
-	}
-
-	it->cur_node = (zone_node_t *)(*trie_it_val(it->tree_it));
+	memset(update, 0, sizeof(*update));
 
 	return KNOT_EOK;
 }
 
-static int iter_get_next_node(zone_update_iter_t *it)
-{
-	trie_it_next(it->tree_it);
-	if (trie_it_finished(it->tree_it)) {
-		trie_it_free(it->tree_it);
-		it->tree_it = NULL;
-		it->cur_node = NULL;
-		return KNOT_ENOENT;
-	}
-
-	it->cur_node = (zone_node_t *)(*trie_it_val(it->tree_it));
-
-	return KNOT_EOK;
-}
-
-static int iter_init(zone_update_iter_t *it, zone_update_t *update, const bool nsec3)
-{
-	memset(it, 0, sizeof(*it));
-
-	it->update = update;
-	it->nsec3 = nsec3;
-	int ret = iter_init_tree_iters(it, update, nsec3);
-	if (ret != KNOT_EOK) {
-		return ret;
-	}
-
-	it->cur_node = (zone_node_t *)(*trie_it_val(it->tree_it));
-
-	return KNOT_EOK;
-}
-
-int zone_update_iter(zone_update_iter_t *it, zone_update_t *update)
-{
-	if (it == NULL || update == NULL) {
-		return KNOT_EINVAL;
-	}
-
-	return iter_init(it, update, false);
-}
-
-int zone_update_iter_nsec3(zone_update_iter_t *it, zone_update_t *update)
-{
-	if (it == NULL || update == NULL) {
-		return KNOT_EINVAL;
-	}
-
-	if (update->flags & UPDATE_FULL) {
-		if (update->new_cont->nsec3_nodes == NULL) {
-			/* No NSEC3 tree. */
-			return KNOT_ENOENT;
-		}
-	} else {
-		if (update->change.add->nsec3_nodes == NULL &&
-		    update->change.remove->nsec3_nodes == NULL) {
-			/* No NSEC3 changes. */
-			return KNOT_ENOENT;
-		}
-	}
-
-	return iter_init(it, update, true);
-}
-
-int zone_update_iter_next(zone_update_iter_t *it)
-{
-	if (it == NULL) {
-		return KNOT_EINVAL;
-	}
-
-	if (it->tree_it != NULL) {
-		int ret = iter_get_next_node(it);
-		if (ret != KNOT_EOK && ret != KNOT_ENOENT) {
-			return ret;
-		}
-	}
-
-	return KNOT_EOK;
-}
-
-const zone_node_t *zone_update_iter_val(zone_update_iter_t *it)
-{
-	if (it != NULL) {
-		return it->cur_node;
-	} else {
-		return NULL;
-	}
-}
-
-void zone_update_iter_finish(zone_update_iter_t *it)
-{
-	if (it == NULL) {
-		return;
-	}
-
-	trie_it_free(it->tree_it);
-}
-
 bool zone_update_no_change(zone_update_t *update)
 {
 	if (update == NULL) {
 		return true;
 	}
 
-	if (update->flags & UPDATE_INCREMENTAL) {
+	if (update->flags & (UPDATE_INCREMENTAL | UPDATE_HYBRID)) {
 		return changeset_empty(&update->change);
 	} else {
 		/* This branch does not make much sense and FULL update will most likely
