@@ -1,4 +1,4 @@
-/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2020 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,16 +11,17 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-*/
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ */
+
 /*!
-* \file
-*
-* \brief Knot error codes.
-*
-* \addtogroup libknot
-* @{
-*/
+ * \file
+ *
+ * \brief Knot error codes.
+ *
+ * \addtogroup libknot
+ * @{
+ */
 
 #pragma once
 
@@ -36,9 +37,20 @@ enum knot_error {
 	KNOT_ENOTSUP       = -ENOTSUP,
 	KNOT_EBUSY         = -EBUSY,
 	KNOT_EAGAIN        = -EAGAIN,
+	KNOT_ENOBUFS       = -ENOBUFS,
+	KNOT_EMFILE        = -EMFILE,
+	KNOT_ENFILE        = -ENFILE,
 	KNOT_EACCES        = -EACCES,
-	KNOT_ECONNREFUSED  = -ECONNREFUSED,
 	KNOT_EISCONN       = -EISCONN,
+	KNOT_ECONNREFUSED  = -ECONNREFUSED,
+	KNOT_EALREADY      = -EALREADY,
+	KNOT_ECONNRESET    = -ECONNRESET,
+	KNOT_ECONNABORTED  = -ECONNABORTED,
+	KNOT_ENETRESET     = -ENETRESET,
+	KNOT_EHOSTUNREACH  = -EHOSTUNREACH,
+	KNOT_ENETUNREACH   = -ENETUNREACH,
+	KNOT_EHOSTDOWN     = -EHOSTDOWN,
+	KNOT_ENETDOWN      = -ENETDOWN,
 	KNOT_EADDRINUSE    = -EADDRINUSE,
 	KNOT_ENOENT        = -ENOENT,
 	KNOT_EEXIST        = -EEXIST,
@@ -62,6 +74,7 @@ enum knot_error {
 	KNOT_ENOZONE,
 	KNOT_ENONODE,
 	KNOT_ENORECORD,
+	KNOT_EISRECORD,
 	KNOT_ENOMASTER,
 	KNOT_EPREREQ,
 	KNOT_ETTL,
@@ -78,6 +91,7 @@ enum knot_error {
 	KNOT_EFILE,
 	KNOT_ESOAINVAL,
 	KNOT_ETRAIL,
+	KNOT_EPROCESSING,
 
 	/* Control states. */
 	KNOT_CTL_ESTOP,
@@ -126,9 +140,6 @@ enum knot_error {
 	KNOT_TXN_EEXISTS,
 	KNOT_TXN_ENOTEXISTS,
 
-	/* Processing error. */
-	KNOT_LAYER_ERROR,
-
 	/* DNSSEC errors. */
 	KNOT_INVALID_PUBLIC_KEY,
 	KNOT_INVALID_PRIVATE_KEY,
@@ -139,6 +150,12 @@ enum knot_error {
 	KNOT_NO_PUBLIC_KEY,
 	KNOT_NO_PRIVATE_KEY,
 
+	/* Another general errors. */
+	KNOT_ELOOP,
+
+	/* Backported. */
+	KNOT_NO_READY_KEY,
+
 	KNOT_ERROR_MAX = -501
 };
 
@@ -167,9 +184,20 @@ inline static int knot_map_errno_code(int code)
 		ERR_ITEM(ENOTSUP),
 		ERR_ITEM(EBUSY),
 		ERR_ITEM(EAGAIN),
+		ERR_ITEM(ENOBUFS),
+		ERR_ITEM(EMFILE),
+		ERR_ITEM(ENFILE),
 		ERR_ITEM(EACCES),
-		ERR_ITEM(ECONNREFUSED),
 		ERR_ITEM(EISCONN),
+		ERR_ITEM(ECONNREFUSED),
+		ERR_ITEM(EALREADY),
+		ERR_ITEM(ECONNRESET),
+		ERR_ITEM(ECONNABORTED),
+		ERR_ITEM(ENETRESET),
+		ERR_ITEM(EHOSTUNREACH),
+		ERR_ITEM(ENETUNREACH),
+		ERR_ITEM(EHOSTDOWN),
+		ERR_ITEM(ENETDOWN),
 		ERR_ITEM(EADDRINUSE),
 		ERR_ITEM(ENOENT),
 		ERR_ITEM(EEXIST),
