@@ -38,9 +38,20 @@ static const struct error errors[] = {
 	{ KNOT_ENOTSUP,       "operation not supported" },
 	{ KNOT_EBUSY,         "requested resource is busy" },
 	{ KNOT_EAGAIN,        "OS lacked necessary resources" },
+	{ KNOT_ENOBUFS,       "no buffers" },
+	{ KNOT_EMFILE,        "too many open files" },
+	{ KNOT_ENFILE,        "too many open files in system" },
 	{ KNOT_EACCES,        "operation not permitted" },
-	{ KNOT_ECONNREFUSED,  "connection refused" },
 	{ KNOT_EISCONN,       "already connected" },
+	{ KNOT_ECONNREFUSED,  "connection refused" },
+	{ KNOT_EALREADY,      "operation already in progress" },
+	{ KNOT_ECONNRESET,    "connection reset by peer" },
+	{ KNOT_ECONNABORTED,  "connection aborted" },
+	{ KNOT_ENETRESET,     "connection aborted by network" },
+	{ KNOT_EHOSTUNREACH,  "host is unreachable" },
+	{ KNOT_ENETUNREACH,   "network is unreachable" },
+	{ KNOT_EHOSTDOWN,     "host is down" },
+	{ KNOT_ENETDOWN,      "network is down" },
 	{ KNOT_EADDRINUSE,    "address already in use" },
 	{ KNOT_ENOENT,        "not exists" },
 	{ KNOT_EEXIST,        "already exists" },
@@ -141,6 +152,9 @@ static const struct error errors[] = {
 	/* Another general errors. */
 	{ KNOT_ELOOP,                 "loop detected" },
 
+	/* Backported. */
+	{ KNOT_NO_READY_KEY,          "no key ready for submission" },
+
 	{ KNOT_ERROR, NULL } /* Terminator */
 };
 
