@@ -1,4 +1,4 @@
-/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2020 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,7 +11,7 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #include <assert.h>
@@ -23,6 +23,7 @@
 #include "libknot/packet/rrset-wire.h"
 #include "libknot/rrtype/naptr.h"
 #include "libknot/rrtype/rrsig.h"
+#include "libknot/rrtype/soa.h"
 #include "contrib/macros.h"
 #include "contrib/mempattern.h"
 #include "contrib/tolower.h"
@@ -329,6 +330,9 @@ static int write_fixed_header(const knot_rrset_t *rrset, uint16_t rrset_index,
 	if ((flags & KNOT_PF_ORIGTTL) && rrset->type == KNOT_RRTYPE_RRSIG) {
 		const knot_rdata_t *rdata = knot_rdataset_at(&rrset->rrs, rrset_index);
 		wire_ctx_write_u32(&write, knot_rrsig_original_ttl(rdata));
+	} else if ((flags & KNOT_PF_SOAMINTTL) && rrset->type == KNOT_RRTYPE_SOA) {
+		const knot_rdata_t *rdata = knot_rdataset_at(&rrset->rrs, rrset_index);
+		wire_ctx_write_u32(&write, MIN(knot_soa_minimum(rdata), rrset->ttl));
 	} else {
 		wire_ctx_write_u32(&write, rrset->ttl);
 	}
@@ -509,14 +513,6 @@ int knot_rrset_to_wire_extra(const knot_rrset_t *rrset, uint8_t *wire,
 	return write - wire;
 }
 
-_public_
-int knot_rrset_to_wire_rotate(const knot_rrset_t *rrset, uint8_t *wire,
-                              uint16_t max_size, uint16_t rotate,
-                              knot_compr_t *compr)
-{
-	return knot_rrset_to_wire_extra(rrset, wire, max_size, rotate, compr, 0);
-}
-
 static int parse_header(const uint8_t *wire, size_t *pos, size_t pkt_size,
                         knot_mm_t *mm, knot_rrset_t *rrset, uint16_t *rdlen)
 {
@@ -533,7 +529,7 @@ static int parse_header(const uint8_t *wire, size_t *pos, size_t pkt_size,
 		return KNOT_EMALF;
 	}
 
-	uint8_t buff[KNOT_DNAME_MAXLEN];
+	knot_dname_storage_t buff;
 	int decompr_size = knot_dname_unpack(buff, src.position, sizeof(buff), wire);
 	if (decompr_size <= 0) {
 		return KNOT_EMALF;
