@@ -1,4 +1,4 @@
-/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2020 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,8 +11,8 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-*/
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ */
 
 #include <arpa/inet.h>
 #include <fcntl.h>
@@ -125,13 +125,24 @@ static int get_addr(const srv_info_t *server,
 	hints.ai_socktype = socktype;
 
 	// Get connection parameters.
-	if (getaddrinfo(server->name, server->service, &hints, info) != 0) {
-		ERR("can't resolve address %s@%s\n",
-		    server->name, server->service);
-		return -1;
+	int ret = getaddrinfo(server->name, server->service, &hints, info);
+	switch (ret) {
+	case 0:
+		return 0;
+#ifdef EAI_ADDRFAMILY	/* EAI_ADDRFAMILY isn't implemented in FreeBSD/macOS anymore. */
+	case EAI_ADDRFAMILY:
+		break;
+#else			/* FreeBSD, macOS, and likely others return EAI_NONAME instead. */
+	case EAI_NONAME:
+		if (iptype != AF_UNSPEC) {
+			break;
+		}
+		/* FALLTHROUGH */
+#endif	/* EAI_ADDRFAMILY */
+	default:
+		ERR("%s for %s@%s\n", gai_strerror(ret), server->name, server->service);
 	}
-
-	return 0;
+	return -1;
 }
 
 void get_addr_str(const struct sockaddr_storage *ss,
@@ -141,7 +152,7 @@ void get_addr_str(const struct sockaddr_storage *ss,
 	char addr_str[SOCKADDR_STRLEN] = {0};
 
 	// Get network address string and port number.
-	sockaddr_tostr(addr_str, sizeof(addr_str), (struct sockaddr *)ss);
+	sockaddr_tostr(addr_str, sizeof(addr_str), ss);
 
 	// Calculate needed buffer size
 	const char *sock_name = get_sockname(socktype);
@@ -212,7 +223,7 @@ int net_init(const srv_info_t    *local,
 	return KNOT_EOK;
 }
 
-/**
+/*!
  * Connect with TCP Fast Open.
  */
 static int fastopen_connect(int sockfd, const struct addrinfo *srv)
@@ -234,7 +245,7 @@ static int fastopen_connect(int sockfd, const struct addrinfo *srv)
 #endif
 }
 
-/**
+/*!
  * Sends data with TCP Fast Open.
  */
 static int fastopen_send(int sockfd, const struct msghdr *msg, int timeout)
@@ -300,6 +311,10 @@ int net_connect(net_t *net)
 			WARN("can't assign address %s\n", net->local->name);
 			return KNOT_NET_ESOCKET;
 		}
+	} else {
+		// Ensure source port is always randomized (even for TCP).
+		struct sockaddr_storage local = { .ss_family = net->srv->ai_family };
+		(void)bind(sockfd, (struct sockaddr *)&local, sockaddr_len(&local));
 	}
 
 	if (net->socktype == SOCK_STREAM) {
@@ -336,7 +351,7 @@ int net_connect(net_t *net)
 
 		// Establish TLS connection.
 		if (net->tls.params != NULL) {
-			int ret = tls_ctx_connect(&net->tls, sockfd, net->tls.params->sni);
+			ret = tls_ctx_connect(&net->tls, sockfd, net->tls.params->sni);
 			if (ret != KNOT_EOK) {
 				close(sockfd);
 				return ret;
