@@ -1,4 +1,4 @@
-/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2019 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,8 +11,8 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-*/
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ */
 
 #include <assert.h>
 #include <errno.h>
@@ -166,7 +166,7 @@ static int parse_full_rr(zs_scanner_t *s, const char* lp)
 
 	/* Class must not differ from specified. */
 	if (s->r_class != s->default_class) {
-		char cls_s[16] = {0};
+		char cls_s[16] = "";
 		knot_rrclass_to_string(s->default_class, cls_s, sizeof(cls_s));
 		ERR("class mismatch '%s'\n", cls_s);
 		return KNOT_EPARSEFAIL;
@@ -371,7 +371,7 @@ static int rr_list_to_packet(knot_pkt_t *dst, list_t *list)
 	assert(dst != NULL);
 	assert(list != NULL);
 
-	ptrnode_t *node = NULL;
+	ptrnode_t *node;
 	WALK_LIST(node, *list) {
 		int ret = knot_pkt_put(dst, KNOT_COMPR_HINT_NONE,
 		                       (knot_rrset_t *)node->d, 0);
@@ -570,7 +570,7 @@ int knsupdate_exec(knsupdate_params_t *params)
 	}
 
 	/* Read from each specified file. */
-	ptrnode_t *n = NULL;
+	ptrnode_t *n;
 	WALK_LIST(n, params->qfiles) {
 		const char *filename = (const char*)n->d;
 		if (strcmp(filename, "-") == 0) {
