@@ -58,6 +58,9 @@ int main(int argc, char *argv[])
 		{ NULL }
 	};
 
+	/* Set the time zone. */
+	tzset();
+
 	/* Parse command line arguments */
 	int opt = 0;
 	while ((opt = getopt_long(argc, argv, "o:t:vVh", opts, NULL)) != -1) {
